from numpy.linalg import solve, norm
import time
import copy
from collections import defaultdict

from django.db import models

from element import *
from qmpy.utils import *
from qmpy.analysis.symmetry import WyckoffSite

class AtomError(Exception):
    '''Atom related problem'''

class SiteError(Exception):
    '''Site related problem'''

class Site(models.Model):
    '''
    Model for a lattice site. 

    Database Attributes:
        structure   : Structure object the site is associated with
        x, y, z     : fractional coordinates of the site (float)
        wyckoff     : WyckoffSite the site is on
        atom_set    : QuerySet of atoms on the site

    Convenience Attributes:
        coord       : equivalent to [x, y, z]
        atoms       : list of Atoms on the site
    '''
    structure = models.ForeignKey('Structure', blank=True, null=True)
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()
    wyckoff = models.ForeignKey(WyckoffSite, blank=True, null=True)

    class Meta:
        app_label = 'qmpy'
        db_table = 'sites'

    # builtins
    def __str__(self):
        if len(self.atoms) == 1:
            return self.atoms[0].element_id
        elif len(self.atoms) > 1:
            elts = [ str(a.element) for a in self.atoms ]
            specs = [ str(a.species) for a in self.atoms ]
            if len(set(elts)) == len(set(specs)):
                return '('+','.join(elts)+')'
            else:
                return '('+','.join(specs)+')'
        else:
            return 'No atom found'

    # django overrides
    def save(self, *args, **kwargs):
        self.structure = self.structure
        for a in self.atoms:
            a.structure = self.structure
            a.site = self
        super(Site, self).save(*args, **kwargs)
        self.atom_set = self.atoms

    _atoms = None
    @property
    def atoms(self):
        if self._atoms is None:
            if self.id is None:
                self._atoms = []
            else:
                self._atoms = list(self.atom_set.all())
        return self._atoms

    @atoms.setter
    def atoms(self, atoms):
        if isinstance(atoms, Atom):
            self._atoms = [atoms]
        elif all([ isinstance(a, Atom) for a in atoms ]):
            self._atoms = atoms
        else:
            raise TypeError('atoms must be a sequence of Atoms')

    @property
    def coord(self):
        return np.array([self.x , self.y, self.z])

    @coord.setter
    def coord(self, values):
        self.x, self.y, self.z = values

    @property
    def cart_coord(self):
        #if self.direct and self.structure is None:
        #    raise SiteError("No associated structure")
        #elif self.direct:
        if self.structure is None:
            raise SiteError("No associated structure")
        return np.dot(self.coord, self.structure.cell)
        #else:
        #    return self.coord

    @property
    def label(self):
        if len(self.atoms) == 1:
            return self.atoms[0].element_id

class Atom(models.Model):
    '''
    Model for an Atom. An atom is defined principally by an element and a
    coordinate, but can also have additional properties like: magnetic moment,
    volume, charge or forces.

    Database Attributes:
        structure
        site
        element
        ox

        x, y, z
        direct
        fx, fy, fz

        magmom
        charge
        volume

    Convenience Attributes:

    Methods:
    '''
    structure = models.ForeignKey('Structure', related_name='atom_set',
                                               null=True)
    site = models.ForeignKey('Site', related_name='atom_set', null=True)

    # species
    element = models.ForeignKey('Element', blank=True, null=True)
    ox = models.IntegerField(default=None, blank=True, null=True)

    # position
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()
    direct = models.BooleanField(default=True)

    # forces
    fx = models.FloatField(blank=True, null=True)
    fy = models.FloatField(blank=True, null=True)
    fz = models.FloatField(blank=True, null=True)

    # properties
    magmom = models.FloatField(blank=True, null=True)
    charge = models.FloatField(blank=True, null=True)
    volume = models.FloatField(blank=True, null=True)

    # symmetry
    occupancy = models.FloatField(default=1)
    wyckoff = models.ForeignKey(WyckoffSite, blank=True, null=True)

    class Meta:
        app_label = 'qmpy'
        db_table = 'atoms'

    # builtins
    def __str__(self):
        return '%s: %8f %8f %8f' % (self.element, 
                self.x, self.y, self.z)

    def __eq__(self, other):
        if self.element_id != other.element_id:
            return False
        if any( abs(x - y) > 1e-3 for x, y in zip(self.coord, other.coord)):
            return False
        return True

    def __cmp__(self, other):
        comp_arr = [[ self.z, other.z ],
                    [ self.y, other.y ],
                    [ self.x, other.x ],
                    [ self.ox, other.ox ],
                    [ self.element_id, other.element_id ]]
        comp_arr = np.array([ row for row in comp_arr if 
                                  all( not x is None for x in row ) ]).T
        if len(comp_arr) == 0:
            return 0
        ind = np.lexsort(comp_arr)
        return ind[0]

    # convenience attributes
    @property
    def forces(self):
        return np.array([self.fx , self.fy, self.fz])

    @forces.setter
    def forces(self, values):
        self.fx, self.fy, self.fz = values

    @property
    def coord(self):
        return np.array([self.x , self.y, self.z])

    @coord.setter
    def coord(self, values):
        self.x, self.y, self.z = values

    @property
    def cart_coord(self):
        if self.direct and self.structure is None:
            raise AtomError("No associated structure")
        elif self.direct:
            return np.dot(self.coord, self.structure.cell)
        else:
            return self.coord

    @property
    def species(self):
        return format_species(self.element_id, self.ox)

    # accessor
    @classmethod
    def get(cls, element, coord, **kwargs):
        '''
        Create a new Atom object. Requires element and coordinate arguments,
        with all other attributes being optional arguments.

        Example:
        >>> Atom.get('Fe', [0,0,0])
        >>> Atom.get('Ni', [0.5, 0.5, 0.5], ox=2, magmom=5, 
        >>>                                 forces=[0.2, 0.2, 0.2],
        >>>                                 volume=101, charge=1.8,
        >>>                                 occupancy=1)
        '''
        atom = Atom()
        atom.element = element
        atom.coord = coord
        valid_keys = ['ox', 'occupancy', 'wyckoff', 'charge', 
                'magmom', 'volume', 'forces']
        for key in valid_keys:
            if key in kwargs:
                setattr(atom, key, kwargs[key])
        return atom

    def copy(self):
        '''
        Creates an exact copy of the atom, only without the matching primary
        key.

        Example:
        >>> a = Atom.get('Fe', [0,0,0])
        >>> a.save()
        >>> a.id
        1
        >>> a.copy()
        >>> a
        <Atom: Fe @ [0.000, 0.000, 0.000]>
        >>> a.id
        None
        
        '''
        atom = Atom()
        atom.element_id = self.element_id
        atom.coord = self.coord
        valid_keys = ['ox', 'occupancy', 'charge', 
                'magmom', 'volume', 'forces']
        for key in valid_keys:
            setattr(atom, key, getattr(self, key))
        return atom

    def as_site(self):
        s = Site()
        s.coord = self.coord
        s.atoms = [self]
        return s

    def is_on(self, site, tol=1e-5):
        vec = self.coord - site.coord
        dist = norm(np.dot(vec - np.round(vec), self.structure.cell))
        return dist < tol



import os, sys
from os.path import dirname, abspath
import matplotlib
matplotlib.use('Agg')

if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "qmpy.db.settings")

    from django.core.management import execute_from_command_line
    
    execute_from_command_line(sys.argv)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from collections import defaultdict
import logging
import gzip
import os
from os.path import exists, isfile, isdir
import time

from django.db import models
import numpy as np

from qmpy.utils import *
import qmpy.materials.structure as st
import qmpy.materials.composition as comp
import qmpy.data.meta_data as ref

def write_poscar(struct, direct=True, distinct_by_ox=False):
    comp = struct.comp
    struct.atoms = atom_sort(struct.atoms)

    cdict = defaultdict(int)
    if not distinct_by_ox:
        ordered_keys = sorted(comp.keys())
        counts = [ int(comp[k]) for k in ordered_keys ]
    else:
        for a in struct.atoms:
            if int(a.oxidation_state) != a.oxidation_state:
                cdict['%s%+f' % (a.element_id, a.oxidation_state)] += 1
            else:
                cdict['%s%+d' % (a.element_id, a.oxidation_state)] += 1
        ordered_keys = sorted([ k for k in cdict.keys() ])
        counts = [ int(cdict[k]) for k in ordered_keys ]

    poscar = ' '.join(set(a.element_id for a in struct.atoms)) + '\n 1.0\n'
    cell = '\n'.join([ ' '.join([ '%f' % v  for v in vec ]) for vec in
        struct.cell ])
    poscar += cell +'\n'
    names = ' '.join( a for a in ordered_keys ) + '\n'
    ntypes = ' '.join( str(n) for n in counts ) + '\n'
    poscar += names
    poscar += ntypes
    if direct:
        poscar += 'direct\n'
        for x,y,z in struct.scaled_coords:
            poscar += ' %.10f %.10f %.10f\n' % (x,y,z)
    else:
        poscar += 'cartesian\n'
        for x, y, z in struct.coords:
            poscar += ' %.10f %.10f %.10f\n' % (x,y,z)
    return poscar

def read_poscar(poscar):
    struct = st.Structure()
    poscar = open(poscar,'r')
    title = poscar.readline().strip()
    scale = float(poscar.readline().strip())
    s = float(scale)
    cell = [[ float(v) for v in poscar.readline().split() ],
            [ float(v) for v in poscar.readline().split() ],
            [ float(v) for v in poscar.readline().split() ]]
    cell = np.array(cell)

    if s > 0:
        struct.cell = cell*s
    else:
        struct.cell = cell
        struct.volume = -1*s

    vasp5 = False
    species = poscar.readline().strip().split()
    try:
        float(species[0])
    except:
        vasp5 = True
        counts = [ int(v) for v in poscar.readline().split() ]

    if not vasp5:
        species = title.strip().split()
        counts = species
        print 'Since not vasp5 format, title MUST be species present'

    atom_types = []
    for n,e in zip(counts, species):
        atom_types += [e]*n

    style = poscar.readline()
    direct = False
    if style[0] in ['D', 'd']:
        direct = True

    struct.natoms = sum(counts)
    struct.ntypes = len(counts)
    composition = defaultdict(float)
    atoms = []
    inv = np.linalg.inv(cell).T
    for i in range(struct.natoms):
        atom = st.Atom()
        atom.element_id = atom_types[i]
        if direct:
            atom.coord = [ float(v) for v in poscar.readline().split() ]
        else:
            cart = [ float(v) for v in poscar.readline().split() ]
            atom.coord = np.dot(inv, cart)
        atom.direct = True
        atoms.append(atom)
        composition[atom_types[i]] += 1
    struct.atoms = atoms
    struct.composition = comp.Composition.get(composition)
    struct.reported_composition = comp.Composition.get(composition)
    return struct



'''
qmpy.data.strings

Contains useful functions for string manipulation. Some important conventions
observed throughout qmpy that are particularly relevent here:
    names are of the form: is FeO2 Ni3B
    formula are of the form: Fe,O2 Ni3,B
    comp are of the form: {'Fe':1, 'O':2} {'Ni':3, 'B':1}
    latex are of the form: FeO_2 Ni_3B

These forms can be converted between with x_to_y functions, where x can be
name, formula, comp or latex. (latex and only be y)

'''
from collections import defaultdict
import itertools
import numpy as np
import re
import yaml
import fractions as frac
import decimal as dec

from qmpy.utils.math import *
from qmpy.data import elements, element_groups

## regex's
re_comp = re.compile('({[^}]*}[,0-9x\.]*|[A-Z][a-wyz]?)([,0-9x\.]*)')
spec_comp = re.compile('([A-Z][a-z]?)([0-9\.]*)([+-]?)')

## Parsing

def parse_comp(value):
    comp = {}
    for elt, amt in re_comp.findall(value):
        if elt in ['D', 'T']:
            elt = 'H'
        if amt == '':
            comp[elt] = 1
        elif is_integer(amt):
            comp[elt] = int(amt)
        else:
            comp[elt] = float(amt)
    return comp

def parse_sitesym(sitesym, sep=','):
    rot = np.zeros((3, 3), dtype='int')
    trans = np.zeros(3)
    for i, s in enumerate (sitesym.split(sep)):
        s = s.lower().strip()
        while s:
            sign = 1
            if s[0] in '+-':
                if s[0] == '-':
                    sign = -1
                s = s[1:]
            if s[0] in 'xyz':
                j = ord(s[0]) - ord('x')
                rot[i, j] = sign
                s = s[1:]
            elif s[0].isdigit() or s[0] == '.':
                n = 0
                while n < len(s) and (s[n].isdigit() or s[n] in '/.'):
                    n += 1
                t = s[:n]
                s = s[n:]
                trans[i] = float(frac.Fraction(t))
            else:
                raise ValueError('Failed to parse symmetry of %s' % (sitesym))
    return rot, trans

def parse_species(value):
    elt, charge, sign = spec_comp.findall(value)[0]
    if charge == '':
        return elt, None
    elif is_integer(charge):
        charge = int(charge)
    else:
        charge = float(charge)

    if sign in ['+', '']:
        return elt, charge
    elif sign == '-':
        return elt, -1*charge

## Formatting

def format_species(element, value):
    if value is None:
        return element
    if is_integer(value):
        ox = str(abs(int(value)))
    else:
        ox = str(abs(float(value))).rstrip('0.')
    name = '%s%s' % (element, ox)
    if value < 0:
        name += '-'
    else:
        name += '+'
    return name

def get_coeffs(values):
    wasdict = False
    if isinstance(values, dict):
        keys, values = zip(*values.items())
        wasdict = True

    coeffs = []
    for v in values:
        if v == 1:
            coeffs.append('')
        elif is_integer(v):
            coeffs.append('%d' % v)
        else:
            coeffs.append(('%.8f' % v).rstrip('0'))

    if wasdict:
        return dict(zip(keys, coeffs))
    else:
        return coeffs

def electronegativity(elt):
    if not elt in elements:
        return 0.0
    else:
        if not 'electronegativity' in elements[elt]:
            return 0.0
        return elements[elt]['electronegativity']

def format_comp(comp, template='{elt}{amt}', join=''):
    elts = sorted(comp.keys(), key=lambda x: electronegativity(x))
    coeffs = get_coeffs(comp)
    return ''.join(template.format(elt=k, amt=coeffs[k]) for k in elts)

def html_comp(comp):
    return format_comp(comp, template='{elt}<sub>{amt}</sub>')

def latex_comp(comp):
    return format_comp(comp, template='{elt}_{{{amt}}}')

def unit_comp(comp):
    tot = float(sum(comp.values()))
    return dict( (k, v/tot) for k, v in comp.items())

def reduce_comp(values):
    '''
    Attempt to construct a 'pretty' composition string. NOTE: Any composition
    within between n-0.05 and n+0.05 will be rounded to n.
    
    Tries 3 options:
        a) Find the greatest common denominator for all coefficients.

        b) If that fails (i.e. the GCD < 1e-10, and there is no accurate
        rational value ), try multiplying the coefficients by the first 20
        prime numbers. For example: if your coefficients are 
        [0.3333333, 0.6666667], this will fail to identify a rational GCD, but
        if you multiply by 3, you get [1, 2]. NOTE: This only works because we
        only keep compositions to 1 decimal place. As a result, 0.6666667
        cannot be properly identified, but 0.6666667*3, can. In fact, [ 0.33,
        0.67 

        c) If None of these work, simply return the literal composition given,
        rounded to 3 decimal places.
    '''

    wasdict = False
    if isinstance(values, dict):
        keys, values = zip(*values.items())
        wasdict = True

    def roundclose(v):
        if is_integer(v, tol=5e-2):
            return int(round(v))
        else:
            return v
    i = 0
    d = 0
    vals = np.array([ roundclose(v) for v in values])
    primes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47]
    while d < 1e-6:
        if i > len(primes)-1:
            vals = np.round(vals*10**3)/10**3
            break
        i += 1
        p = primes[i-1]
        d = reduce(frac.gcd, [ roundclose(v) for v in vals*p ])
    else:
        vals = np.round(vals*p/d)

    if wasdict:
        return dict(zip(keys, vals))
    else:
        return vals.tolist()

def normalize_comp(values):
    wasdict = False
    if isinstance(values, dict):
        keys, values = zip(*values.items())
        wasdict = True

    vals = np.array(values)
    vals /= min(vals)

    if wasdict:
        return dict(zip(keys, vals))
    else:
        return vals.tolist()

def parse_formula_regex(formula):
    '''Take in a generalized expression for a composition. Use symbols for
    groups of elements for combinatorial replacements.

    Groups of elements can be identified by:
        - valence shell: 3d, 5s, 2p, etc...
        - group: G1 (alkali), G17 (halides), G4 (2d2), etc...
        - row: R1, R3
        - block: DD (all d-block), PP (all p-block), etc...

    Examples:
    >>> parse_formula_regex('Fe2O3')
    [{'Fe':2, 'O':3}]
    >>> parse_formula_regex('{Fe,Ni}2O3')
    [{'Fe':2, 'O':3}, {'Ni':2, 'O':3}]
    >>> parse_formula_regex('{3d}2O3')
    [{'Co': 2.0, 'O': 3.0}, {'Cr': 2.0, 'O': 3.0}, {'Cu': 2.0, 'O':3.0}, 
    {'Fe': 2.0, 'O': 3.0}, {'Mn': 2.0, 'O': 3.0}, {'Ni': 2.0, 'O': 3.0},
    {'Sc': 2.0, 'O': 3.0}, {'O': 3.0, 'Ti': 2.0}, {'O': 3.0, 'V': 2.0}, 
    {'Zn':2.0, 'O': 3.0}]

    '''
    formulae = []
    sfind = re.compile('({[^}]*}[,0-9x\.]*|[A-Z][a-wyz]?[,0-9x\.]*)')
    matches = sfind.findall(formula)
    for term in matches:
        if '{' in term:
            symbols, amt = term.replace('{','').split('}')
            symbols = symbols.split(',')
            elts = []
            for symbol in symbols:
                if symbol in element_groups.keys():
                    elts += [ e for e in element_groups[symbol] ]
                else:
                    elts += [ symbol ]
            formulae.append([ e+amt for e in elts ])
        else:
            formulae.append([term])
    return [ parse_comp(''.join(ref)) for ref in itertools.product(*formulae) ]



from django.db import models
import yaml
import os.path

import numpy as np

from qmpy.db.custom import NumpyArrayField

#class Orbital(models.Model):
#    name = models.CharField(max_length=10)

#class EnergyLevel(models.Model):
#    dos = models.ForeignKey('DOS', related_name='energy_levels')
#    energy = models.FloatField(blank=True, null=True)
#    occupation = models.FloatField(blank=True, null=True)
#    orbital = models.ForiegnKey(Orbital)
#
#    class Meta:
#        db_table = 'states'
#        app_label = 'qmpy'
#        order_by = 'energy'

class DOS(models.Model):
    '''
    Density of states model.

    Database attributes:
        efermi      : fermi energy (float)
        band_gap    : band gap (float)
        vbm         : valence band minimum (float)
        cbm         : conduction band minimum (float)
        file        : DOSCAR file location (string)

    '''

    entry = models.ForeignKey('Entry', null=True)
    efermi = models.FloatField(default=0.0)
    gap = models.FloatField(blank=True, null=True)
    data = NumpyArrayField(blank=True, null=True)
    file = models.CharField(max_length=128, blank=True, null=True)

    class Meta:
        db_table = 'dos'
        app_label = 'qmpy'

    def save(self, *args, **kwargs):
        super(DOS, self).save(*args, **kwargs)

    @classmethod
    def read(cls, doscar='DOSCAR', efermi=0.0):
        dos = DOS(file=os.path.abspath(doscar))
        dos._efermi = 0.0
        dos.read_doscar(dos.file)
        dos.efermi = efermi
        return dos

    @property
    def efermi(self):
        return self._efermi

    @efermi.setter
    def efermi(self, efermi):
        """Set the Fermi level."""
        ef = efermi - self._efermi
        self._efermi = efermi
        try:
            self.data[0, :] = self.data[0, :] + ef
            self._site_dos[:, 0, :] = self._site_dos[:, 0, :] + ef
        except IndexError:
            pass

    @property
    def energy(self):
        """Return the array with the energies."""
        return self.data[0, :]

    def find_vbm(self, tol=1e-3):
        i = (np.abs(self.energy)).argmin()
        if self.energy[i] > 0:
            i -= 1
        edos = self.total_dos
        while edos[i] < tol:
            i -= 1
        else:
            self.vbm = self.energy[i]
        return self.vbm

    def find_cbm(self, tol=1e-3):
        i = (np.abs(self.energy)).argmin()
        if self.energy[i] < 0:
            i += 1
        edos = self.total_dos
        while edos[i] < tol:
            i += 1
        else:
            self.cbm = self.energy[i]
        return self.cbm

    def find_gap(self, tol=1e-3):
        i0 = (np.abs(self.energy)).argmin()
        i1 = ( i0-1 if self.energy[i0] > 0 else i0+1 )
        edos = self.total_dos
        if edos[i0] > tol and edos[i1] > tol:
            self.gap = 0
        else:
            self.gap = self.find_cbm(tol) - self.find_vbm(tol)
        return self.gap

    def jqplot_dos(self):
        d1 = '[ '
        for energy, dos in zip(self.energy, self.total_dos):
            d1 += ' [ %s, %s ],' % (energy, dos)
        return d1.rstrip(',') + ' ]'

    def jqplot_efermi(self):
        return '[ [%s, %s], [%s, %s] ]' % (0, 0, 0, max(self.total_dos)*1.1 )

    def jqplot_ymax(self):
        return max(self.total_dos)*1.1
            
    def site_dos(self, atom, orbital):
        """Return an NDOSx1 array with dos for the chosen atom and orbital.

        atom: int
            Atom index
        orbital: int or str
            Which orbital to plot

        If the orbital is given as an integer:
        If spin-unpolarized calculation, no phase factors:
        s = 0, p = 1, d = 2
        Spin-polarized, no phase factors:
        s-up = 0, s-down = 1, p-up = 2, p-down = 3, d-up = 4, d-down = 5
        If phase factors have been calculated, orbitals are
        s, py, pz, px, dxy, dyz, dz2, dxz, dx2
        double in the above fashion if spin polarized.

        """
        # Integer indexing for orbitals starts from 1 in the _site_dos array
        # since the 0th column contains the energies
        if isinstance(orbital, int):
            return self._site_dos[atom, orbital + 1, :]
        n = self._site_dos.shape[1]
        if n == 4:
            norb = {'s':1, 'p':2, 'd':3}
        elif n == 7:
            norb = {'s+':1, 's-up':1, 's-':2, 's-down':2,
                    'p+':3, 'p-up':3, 'p-':4, 'p-down':4,
                    'd+':5, 'd-up':5, 'd-':6, 'd-down':6}
        elif n == 10:
            norb = {'s':1, 'py':2, 'pz':3, 'px':4,
                    'dxy':5, 'dyz':6, 'dz2':7, 'dxz':8,
                    'dx2':9}
        elif n == 19:
            norb = {'s+':1, 's-up':1, 's-':2, 's-down':2,
                    'py+':3, 'py-up':3, 'py-':4, 'py-down':4,
                    'pz+':5, 'pz-up':5, 'pz-':6, 'pz-down':6,
                    'px+':7, 'px-up':7, 'px-':8, 'px-down':8,
                    'dxy+':9, 'dxy-up':9, 'dxy-':10, 'dxy-down':10,
                    'dyz+':11, 'dyz-up':11, 'dyz-':12, 'dyz-down':12,
                    'dz2+':13, 'dz2-up':13, 'dz2-':14, 'dz2-down':14,
                    'dxz+':15, 'dxz-up':15, 'dxz-':16, 'dxz-down':16,
                    'dx2+':17, 'dx2-up':17, 'dx2-':18, 'dx2-down':18}
        return self._site_dos[atom, norb[orbital.lower()], :]

    @property
    def dos(self):
        if self.data.shape[0] == 3:
            return self.data[1, :]
        elif self.data.shape[0] == 5:
            return self.data[1:3, :]

    @property
    def total_dos(self):
        if self.data.shape[0] == 3:
            return self.data[1, :]
        elif self.data.shape[0] == 5:
            return np.sum(self.data[1:3, :], 0)

    @property
    def integrated_dos(self):
        if self.data.shape[0] == 3:
            return self.data[2, :]
        elif self.data.shape[0] == 5:
            return self.data[3:5, :]

    def read_doscar(self, fname="DOSCAR"):
        """Read a VASP DOSCAR file"""
        if os.path.getsize(fname) < 300:
            return
        f = open(fname)
        natoms = int(f.readline().split()[0])
        [f.readline() for nn in range(4)]  # Skip next 4 lines.
        # First we have a block with total and total integrated DOS
        ndos, efermi = f.readline().split()[2:4]
        self._efermi = float(efermi)
        ndos = int(ndos)
        dos = []
        for nd in xrange(ndos):
            dos.append(np.array([float(x) for x in f.readline().split()]))
        self.data = np.array(dos).T
        # Next we have one block per atom, if INCAR contains the stuff
        # necessary for generating site-projected DOS
        dos = []
        for na in xrange(natoms):
            line = f.readline()
            if line == '':
                # No site-projected DOS
                break
            ndos = int(line.split()[2])
            line = f.readline().split()
            cdos = np.empty((ndos, len(line)))
            cdos[0] = np.array(line)
            for nd in xrange(1, ndos):
                line = f.readline().split()
                cdos[nd] = np.array([float(x) for x in line])
            dos.append(cdos.T)
        self._site_dos = np.array(dos)



from qmpy.data import elements

pressure_words = [
        'pressure',
        'anvil cell',
        'dac',
        'mpa', 'gpa',
        'mbar', 'kbar']


def is_likely_high_pressure(structure):
    if structure.pressure > 102:
        return True
    if structure.reference:
        if structure.reference.title:
            title = structure.reference.title.lower()
            for w in pressure_words:
                if w in title:
                    return True
    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from django.db import models
import json

from qmpy.materials.composition import Composition

from utils import *

class ExperimentalFormationEnergy(models.Model):
    composition = models.ForeignKey('Composition', null=True, blank=True)
    delta_e = models.FloatField(null=True)
    delta_g = models.FloatField(null=True)
    source = models.CharField(max_length=127, blank=True, null=True)
    dft = models.BooleanField(default=False)

    class Meta:
        app_label = 'qmpy'
        db_table = 'experimental_formation_energies'

    def __str__(self):
        return '%s : %s' % (self.composition, self.delta_e)

    @classmethod
    def read_file(cls, filename, dft=False):
        source = filename.split('.')[0]
        expts = []
        for line in open(filename, 'r'):
            comp, energy = line.strip().split()
            print comp
            expt, new = ExperimentalFormationEnergy.objects.get_or_create(delta_e=energy,
                    composition=Composition.get(comp),
                    source=source, 
                    dft=dft)
            if new:
                expt.save()
            expts.append(expt)
        return expts

class Author(models.Model):
    last = models.CharField(max_length=30, blank=True, null=True)
    first = models.CharField(max_length=30, blank=True, null=True)
    class Meta:
        app_label = 'qmpy'
        db_table = 'authors'

    @property
    def proper_last(self):
        if not self.last:
            return None
        elif len(self.last) == 1:
            return self.last.upper()
        else:
            return self.last[0].upper()+self.last[1:]

    @property
    def proper_first(self):
        if not self.first:
            return None
        elif len(self.first.split()) == 1:
            if len(self.first) == 1:
                return self.first[0].upper()+'.'
            else:
                return self.first[0].upper()+self.first[1:]
        else:
            return ' '.join(l[0].upper()+"." for l in self.first.split())

    def __str__(self):
        if self.first and self.last:
            return '%s, %s' % (self.proper_last, self.proper_first)
        elif self.first and not self.last:
            return self.proper_first
        elif self.last and not self.first:
            return self.proper_last

    @classmethod
    def from_name(cls, name):
        commas = name.count(',')
        spaces = name.count(' ')
        if commas == 1:
            last, first = name.split(',')
        elif commas == 0 and spaces:
            first = name.split()[0]
            last = ' '.join(name.split()[1:])
        else:
            last = name.split(',')[0]
            first = ' '.join(name.split(',')[1:])
        author, new = Author.objects.get_or_create(
                last=last.lower().strip(), first=first.lower().strip())
        return author


class Journal(models.Model):
    code = models.CharField(max_length=10, unique=True)
    name = models.TextField(null=True)
    class Meta:
        app_label = 'qmpy'
        db_table = 'journals'

    def __str__(self):
        if self.name:
            return self.name
        elif self.code:
            return self.code

class Reference(models.Model):
    author_set = models.ManyToManyField(Author, related_name='references',
            null=True)
    journal = models.ForeignKey(Journal, related_name='references', 
            null=True)
    title = models.TextField(null=True)
    volume = models.IntegerField(null=True)
    page_first = models.IntegerField(null=True)
    page_last = models.IntegerField(null=True)
    year = models.IntegerField(null=True)
    volume = models.IntegerField(null=True)

    _authors = []
    class Meta:
        app_label = 'qmpy'
        db_table = 'publications'

    def save(self,*args, **kwargs):
        if self.journal:
            self.journal.save()
        super(Reference, self).save(*args, **kwargs)
        self.author_set = self._authors

    @property
    def authors(self):
        if not self._authors:
            self._authors = list(self.author_set.all())
        return self._authors

    @authors.setter
    def authors(self, authors):
        self._authors = authors

    def __str__(self):
        s = self.title
        if self.authors:
            s += ': %s' % self.authors[0]
        if self.journal is not None:
            s += ': %s' % self.journal
        return s

    @property
    def citation(self):
        retval = ', '.join(str(a) for a in self.authors)
        if self.year:
            retval += '('+str(self.year)+')'
        if self.title:
            retval += '. '+self.title.strip().rstrip('.')
        if self.journal:
            retval += '. '+str(self.journal)
            if self.volume:
                retval += ', '+str(self.volume)
                if self.page_first:
                    retval += ', '+str(self.page_first)
                    if self.page_last:
                        retval += '-'+str(self.page_last)
        return retval+'.'




import numpy as np
from qmpy.utils import *
from composition import Composition

def strain(x1, x2=None, x3=None):
    raise NotImplementedError

def shear():
    raise NotImplementedError

def displace_atom():
    raise NotImplementedError

def joggle_atoms(structure):
    raise NotImplementedError

def replace(structure, replacements):
    raise NotImplementedError

def lattice_transform(structure, transform):
    transform = np.array(transform)
    if transform.shape == (3,3):
        print 'is a full lattice transform matrix'
    elif transform.shape in [ (1,3), (3,1) ]:
        transform = p.eye(3)*transform

    cell = np.dot(structure.cell, transform)
    raise NotImplementedError

def substitute(structure, replace, rescale=True):
    '''Replace atoms, as specified in a dict of pairs. 
    Optionally, can rescale the new volume to the total volume of the
    elemental phases of the new composition.

    Example:
    >>> s = Structure.create('POSCAR-Fe2O3')
    >>> s.substitute({'Fe':'Ni', 'O':'F'} rescale=True)
    '''

    struct = structure.copy()
    volume = 0.0
    for atom in struct:
        if atom.element_id in replace:
            atom.element_id = replace[atom.element_id]
        volume += atom.element.volume
    if rescale:
        struct.volume = volume
    struct.composition = Composition.get(struct.comp)
    return struct




from django.db import models

from qmpy.data.meta_data import *

class XRD(models.Model):
    identifier = models.CharField(max_length=100, db_index=True)
    wavelength = models.FloatField(blank=True, null=True)
    source = models.CharField(max_length=128)
    notes = models.TextField(blank=True)

    class Meta:
        db_table = 'xrd'
        app_label = 'qmpy'

class Peak(models.Model):
    xrd = models.ForeignKey(XRD, related_name='peak_set')
    angle = models.FloatField()
    intensity = models.FloatField(blank=True, null=True)
    width = models.FloatField(blank=True, null=True)

    class Meta:
        db_table = 'peaks'
        app_label = 'qmpy'



from collections import defaultdict
from django.db import models
import cPickle
import numpy as np
import ast

class TagField(models.TextField):
    description = "Stores tags in a single database column."
    __metaclass__ = models.SubfieldBase

    def __init__(self, delimiter="|", *args, **kwargs):
        self.delimiter = delimiter
        super(TagField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        if isinstance(value, list):
            return value

        if not value:
            return []
        # Otherwise, split by delimiter
        return value.split(self.delimiter)

    def get_prep_value(self, value):
        return self.delimiter.join(value)

class NumpyArrayField(models.TextField):
    description = "Stores a Numpy ndarray."
    __metaclass__ = models.SubfieldBase

    def __init__(self, *args, **kwargs):
        super(NumpyArrayField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        if isinstance(value, list):
            value = np.array(list)
        if isinstance(value, np.ndarray):
            return value

        if not value:
            return np.array([])
        return np.array(cPickle.loads(str(value)))

    def get_prep_value(self, value):
        if isinstance(value, list):
            return cPickle.dumps(value)
        elif isinstance(value, np.ndarray):
            return cPickle.dumps(value.tolist())
        else:
            raise TypeError('%s is not a list or numpy array' % value)

class DictField(models.TextField):
    __metaclass__ = models.SubfieldBase
    description = "Stores a python dictionary"

    def __init__(self, *args, **kwargs):
        super(DictField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        if not value:
            value = {}

        if isinstance(value, dict):
            return value

        return ast.literal_eval(value)

    def get_prep_value(self, value):
        if value is None:
            return value

        if isinstance(value, defaultdict):
            value = dict(value)

        return unicode(value)

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_db_prep_value(value)

class DictModel(models.Model):
    data = DictField(max_length=255, primary_key=True)

    class Meta:
        abstract = True

    ## dict methods
    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()




import json
import os.path
import time
import sys
import logging
from datetime import datetime, timedelta

import django.db as ddb

from qmpy.computing.queue import Task, Job
from qmpy.computing.resources import *
from qmpy.utils.daemon import Daemon
from qmpy.analysis.vasp.calculation import VaspError

logger = logging.getLogger(__name__)

def check_die():
    if os.path.exists('/home/sjk648/die'):
        exit(0)

class JobManager(Daemon):
    '''
    Class to manage collection of running Job objects. When running, it
    collects any Job whose is_done method evaluates to True. While the manager
    is running (as a Daemon or normally) it will look for a file named "die" in
    the home directory of the user who initialized the manager, and if found it
    will stop cleanly.

    '''
    def run(self):
        os.umask(022)
        while True:
            if len(ddb.connection.queries) > 10000:
                ddb.connection.queries = []
            jobs = Job.objects.filter(state=1, account__host__state=1,
                    created__lt=datetime.now() - timedelta(seconds=-300))
            for job in jobs:    
                check_die()
                if job.is_done():
                    print job
                    job.collect()
                    job.save()
            for i in range(20):
                time.sleep(1)
                check_die()

class TaskManager(Daemon):
    '''
    Class to manage creation of Job objects from uncompleted Tasks. While
    running it finds the next available Task (based on additional keyword
    arguments) and attempts to submit all Jobs return by the tasks get_jobs()
    method. While the manager is running (as a Daemon or normally) it will 
    look for a file named "die" in the home directory of the user who 
    initialized the manager, and if found it will stop cleanly.

    Conditions for Task eligibility are:
        - A resource assosiated with the Task (via its projects list) has CPUs
          available.
        - if the manager is called with the project keyword, only Tasks
          associated with that project will be run
        - if the manager is called with the path keyword, only Tasks whose
          entry contain the path or path fragment will be run.
    '''
    def run(self, project=None, path=None, id=None):
        os.umask(022)
        while True:
            if len(ddb.connection.queries) > 10000:
                ddb.connection.queries = []
            check_die()

            tasks = Task.objects.none()
            while not tasks.exists():
                if project:
                    tasks = Project.get(project).task_set.filter(state=0)
                else:
                    tasks = Task.objects.filter(project_set__state=1, state=0)
                tasks = tasks.exclude(entry__meta_data__type='hold')
                if tasks.exists():
                    break

                Project.objects.filter(state=0).update(state=1)
                for i in range(120):
                    time.sleep(1)
                    check_die()

            time.sleep(10)
            for task in tasks.order_by('priority')[:50]:
                check_die()
                if not task.eligible_to_run:
                    task.state = -1
                    task.save()
                    continue

                print task

                try:
                    jobs = task.get_jobs()
                except VaspError:
                    print '    -VASP Error.'
                    task.state = -1
                    task.save()
                    continue
                except Exception, err:
                    task.state = -1
                    task.save()
                    print '   -Unknown error!', err
                    continue

                for job in jobs:
                    while not job.state == 1:
                        try:
                            job.submit()
                        except Exception:
                            print '  -waiting 30 seconds to resubmit'
                            time.sleep(30)
                    job.save()
                task.save()
                task.entry.save()





# thermopy.visualize

from qmpy.utils import *
import networkx as nx
import sys
if not 'matplotlib' in sys.modules:
    import matplotlib
    try:
        import backend_wx
        matplotlib.use('WXAgg')
    except ImportError:
        matplotlib.use('Agg')
import matplotlib.pylab as plt

def phase_diagram(space, unstable=False, axes=None, **kwargs):
    '''
    Makes an xwindow with an appropriate phase diagram

    Example:
    >>> space = PhaseSpace('Fe-Li-O')
    >>> space.phase_diagram()
    '''
    if len(space.bounds) == 1:
        raise NotImplementedError
    elif len(space.bounds) == 2:
        hull_plot(space, unstable=unstable, axes=axes, **kwargs)
    elif len(space.bounds) == 3:
        gtri_plot(space, unstable=unstable, axes=axes, **kwargs)
    elif len(space.bounds) == 4:
        gtet_plot(space, unstable=unstable, axes=axes, **kwargs)
    else:
        graph_plot(space, unstable=unstable, axes=axes, **kwargs)

def print_script(space, unstable=False, axes=None, **kwargs):
    '''
    Returns a string which is a pyhon script that generates a phase 
    diagram. Useful for moving labels around to make a graph look better

    Example:
    >>> space = PhaseSpace('Fe-Li-O')
    >>> space.print_script()
    '''

    if len(space.bounds) == 1:
        raise NotImplementedError
    elif len(space.bounds) == 2:
        hull_script(space, unstable=unstable, axes=None, **kwargs)
    elif len(space.bounds) == 3:
        gtri_script(space, unstable=unstable, axes=None, **kwargs)
    elif len(space.bounds) == 4:
        gtet_script(space, unstable=unstable, axes=None, **kwargs)
    else:
        graph_script(space, unstable=unstable, axes=None, **kwargs)

def hull_plot(space, axes=None, unstable=False, **kwargs):
    if axes is None:
        fig = plt.figure()
        axes = fig.add_subplot(111)

    axes.set_xlim([0,1])

    plt.ylabel('$\\rm{\Delta H [eV/atom]}$')
    plt.xlabel('$\\rm{%s_{x}%s_{1-x}}$' % (
        format_comp(space.bounds[0]),
        format_comp(space.bounds[1])))
    
    #plot tie lines
    for p1, p2 in space.tie_lines:
        c1 = space.coord(p1)
        c2 = space.coord(p2)
        if abs(c1[1]) < 1e-6 or abs(c2[1]) < 1e-6:
            if abs(c1[1] - 1) < 1e-6 or abs(c2[1] - 1) < 1e-6:
                if len(space.tie_lines) > 1:
                    continue
        axes.plot([c1[1],c2[1]], [p1.energy, p2.energy], 'k')

    #plot compounds
    for phase in space.stable:
        if not space.in_bounds(phase):
            continue
        x = space.coord(phase.unit_comp)[1]
        axes.plot(x, phase.energy, 'ro')
        axes.annotate('$\\rm{%s}$' % phase.latex, xy=(x, phase.energy))

    if unstable:
        for phase in space.unstable:
            if not space.in_bounds(phase):
                continue
            x = space.coord(phase.unit_comp)[1]
            axes.plot(x, phase.energy, 'bo')
            axes.annotate('$\\rm{%s}$' % phase.latex, xy=(x, phase.energy))
    if axes is None:
        plt.show()

def hull_script(space, axes=None, unstable=False, **kwargs):
    print 'import matplotlib'
    print 'matplotlib.rcParams.update({"font.size": 16})'
    print 'import matplotlib.pylab as plt'
    print
    print 'fig = plt.figure()'
    print 'axes = fig.add_subplot(111)'
    print 'axes.set_xlim([0,1])'
    print 'plt.ylabel("$\\\\rm{\\Delta H [eV/atom]}$")'
    print 'plt.xlabel("$\\\\rm{%s_{x}%s_{1-x}}$")' % (
            format_comp(space.bounds[0]),
            format_comp(space.bounds[1]))
    print
    for p1, p2 in space.tie_lines:
        c1 = space.coord(p1)
        c2 = space.coord(p2)
        if abs(c1[1]) < 1e-6 or abs(c2[1]) < 1e-6:
            if abs(c1[1] - 1) < 1e-6 or abs(c2[1] - 1) < 1e-6:
                if len(space.tie_lines) > 1:
                    continue
        print 'axes.plot([%s, %s], [%s, %s], "k")' % (
                c1[1], c2[1], p1.energy, p2.energy )

    for p in space.stable:
        if not space.in_bounds(p):
            continue
        x = space.coord(p.unit_comp)[1]
        print 'axes.plot(%s, %s, "ro")' % (
                x, p.energy)
        print 'axes.annotate( "$\\\\rm{%s}$", xy=( %s, %s) )' % (
                p.latex, x, p.energy )

def gtri_plot(space, axes=None, unstable=False, labels=True, **kwargs):
    if axes is None:
        fig = plt.figure()
        axes = fig.add_subplot(111)

    axes.axis([-0.1,1.1,-0.1,1.1])
    axes.axis('off')
    axes.set_aspect('equal')

    #plot lines
    for p1, p2 in space.tie_lines:
        x1, y1 = coord_to_gtri(space.coord(p1))
        x2, y2 = coord_to_gtri(space.coord(p2))
        axes.plot([x1, x2], [y1, y2], 'k')

    #plot compounds 
    for p in space.stable:
        if not space.in_bounds(p):
            continue
        x,y = coord_to_gtri(space.coord(p))
        axes.plot(x, y, 'ro')
        if labels:
            axes.annotate('$\\rm{%s}$' % p.latex, xy=(x, y), color='red')

    if unstable:
        for p in space.unstable:
            if not space.in_bounds(p):
                continue
            x,y = coord_to_gtri(space.coord(p))
            axes.plot(x, y, 'bo')
            if labels:
                axes.annotate('$\\rm{%s}$' % p.latex, xy=(x, y))
    if axes is None:
        plt.show()

def gtri_script(space, axes=None, unstable=False, labels=True, **kwargs):
    print 'import matplotlib.pylab as plt'
    print
    print 'axes = plt.axes()'
    print 'axes.axis([-0.1,1.1,-0.1,1.1])'
    print 'axes.axis("off")'
    print 'axes.set_aspect("equal")'
    print
    
    #plot lines
    for p1, p2 in space.tie_lines:
        x1, y1 = coord_to_gtri(space.coord(p1))
        x2, y2 = coord_to_gtri(space.coord(p2))
        print 'axes.plot([%s, %s], [%s, %s], "k")' % (
                x1, x2, y1, y2)

    #plot compounds 
    for p in space.stable:
        if not space.in_bounds(p):
            continue
        x,y = coord_to_gtri(space.coord(p))
        print 'axes.plot(%s, %s, "ro")' % (x, y)
        if labels:
            print 'axes.annotate("$\\\\rm{%s}$" , xy=(%s, %s))' % (
                    p.latex, x, y)

    if unstable:
        for p in space.unstable:
            if not space.in_bounds(p):
                continue
            x,y = coord_to_gtri(space.coord(p))
            print 'axes.plot(%s, %s, "bo")' % (x, y)
            if labels:
                print 'axes.annotate("$\\rm{%s}$", xy=(%s, %s))' % (p.latex, x, y)

def gtet_plot(space, axes=None, unstable=False, labels=True, **kwargs):
    if axes is None:
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        axes = fig.gca(projection='3d')

    axes.axis('off')
    
    #plot lines
    for p1, p2 in space.tie_lines:
        x1, y1, z1 = coord_to_gtet(space.coord(p1))
        x2, y2, z2 = coord_to_gtet(space.coord(p2))
        axes.plot([x1, x2], [y1, y2], [z1, z2], 'k')

    #plot compounds
    for p in space.stable:
        if not space.in_bounds(p):
            continue
        x,y,z = coord_to_gtet(space.coord(p))
        axes.plot([x], [y], [z], 'ro')
        if labels:
            axes.text( x, y, z, '$\\rm{%s}$' % p.latex)

    if unstable:
        for p in space.unstable:
            if not space.in_bounds(p):
                continue
            x,y,z = coord_to_gtet(space.coord(p))
            axes.plot([x], [y], [z], 'bo')   
            if labels:
                axes.text( x, y, z, '$\\rm{%s}$' % p.latex)
    if axes is None:
        plt.show()

def graph_plot(space, axes=None, unstable=False, labels=True, **kwargs):
    G = space.graph
    if axes is None:
        fig = plt.figure()
        axes = fig.add_subplot(111)

    nx.draw(G, ax=axes, node_size=25, edge_color='k',
            with_labels=labels, 
            **kwargs)

    if axes is None:
        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import copy
import json
import time
import gzip
import numpy as np
import numpy.linalg
import logging
from os.path import exists, isfile, isdir

from django.db import models

from qmpy.materials.structure import *
from qmpy.io.poscar import write_poscar, read_poscar
from potential import *
from qmpy.data import chem_pots
from qmpy.db.custom import DictField, NumpyArrayField
import qmpy.utils as utils
import qmpy.db.custom as cdb
from qmpy.configuration.vasp_settings import *
import qmpy.analysis.thermodynamics as thermo
import dos
from qmpy.data.meta_data import MetaData

logger = logging.getLogger(__name__)

def value_formatter(value):
    if isinstance(value, list):
        return ' '.join(map(value_formatter, value))
    elif isinstance(value, basestring):
        return value.upper()
    elif isinstance(value, bool):
        return ('.%s.' % value).upper()
    elif isinstance(value, int):
        return str(value)
    elif isinstance(value, float):
        return '%0.8g' % value
    else:
        return str(value)

def vasp_format(key, value):
    return ' %s = %s' % (key.upper(), value_formatter(value))

class VaspError(Exception):
    '''General problem with vasp calculation.'''

@add_meta_data('error')
class Calculation(models.Model):
    #= labeling =#
    configuration = models.CharField(db_index=True, max_length=15, 
            null=True, blank=True)
    meta_data = models.ManyToManyField(MetaData)

    label = models.CharField(max_length=63, default='')
    entry = models.ForeignKey('Entry', db_index=True, null=True, blank=True)
    path = models.CharField(max_length=255, null=True, db_index=True)

    composition = models.ForeignKey('Composition', null=True, blank=True)
    element_set = models.ManyToManyField('Element')
    natoms = models.IntegerField(blank=True, null=True)

    #= inputs =#
    input = models.ForeignKey(Structure, 
                              related_name='calculated',
                              null=True, blank=True)
    hubbard_set = models.ManyToManyField('Hubbard')
    potential_set = models.ManyToManyField('Potential')
    settings = DictField(blank=True, null=True)
    
    #= outputs =#
    output = models.ForeignKey(Structure, 
                               related_name='source',
                               null=True, blank=True)

    energy = models.FloatField(null=True, blank=True)
    energy_pa = models.FloatField(null=True, blank=True)
    magmom = models.FloatField(blank=True, null=True)
    magmom_pa = models.FloatField(blank=True, null=True)
    dos = models.ForeignKey('DOS', blank=True, null=True)
    band_gap = models.FloatField(blank=True, null=True)
    irreducible_kpoints = models.FloatField(blank=True, null=True)

    #= progress/completion =#
    attempt = models.IntegerField(default=0, blank=True, null=True)
    nsteps = models.IntegerField(blank=True, null=True)
    converged = models.NullBooleanField(null=True)
    runtime = models.FloatField(blank=True, null=True)

    #= other thermodata =#
    delta_e = models.FloatField(blank=True, null=True)
    stability = models.FloatField(blank=True, null=True)

    #= Non-stored values =#
    outcar = None
    kpoints = None
    occupations = None

    class Meta:
        app_label = 'qmpy'
        db_table = 'calculations'

    # builtins
    def __str__(self):
        retval = ''
        if self.input:
            retval += self.input.name+' @ '
        if self.configuration:
            retval += self.configuration+' settings'
        elif 'prec' in self.settings:
            retval += 'PREC='+self.settings['prec'].upper()
        if self.settings.get('nsw', 1) <= 1:
            retval += ', static'
        if not retval:
            return 'Blank'
        return retval

    # django overrides
    def save(self, *args, **kwargs):
        if self.output is not None:
            self.output.entry = self.entry
            self.output.save()
            self.output_id = self.output.id
            self.composition = self.output.composition
        if self.input is not None:
            self.input.entry = self.entry
            self.input.save()
            self.input_id = self.input.id
            self.composition = self.input.composition
        if self.dos is not None:
            self.dos.entry = self.entry
            self.dos.save()
            self.dos_id = self.dos.id
        super(Calculation, self).save(*args, **kwargs)
        self.hubbard_set = self.hubbards
        self.potential_set = self.potentials
        self.element_set = self.elements
        self.meta_data = self.error_objects

    # django caching
    _potentials = None
    @property
    def potentials(self):
        if self._potentials is None:
            if not self.id:
                self._potentials = []
            else:
                self._potentials = list(self.potential_set.all())
        return self._potentials

    @potentials.setter
    def potentials(self, potentials):
        self._potentials = potentials

    _elements = None
    @property
    def elements(self):
        if self._elements is None:
            if self.id:
                self._elements = list(self.element_set.all())
            else:
                self._elements = list(set([ a.element for a in self.input ]))
        return self._elements

    @elements.setter
    def elements(self, elements):
        self._elements = elements

    _hubbards = None
    @property
    def hubbards(self):
        if self._hubbards is None:
            if not self.id:
                self._hubbards = []
            else:
                self._hubbards = list(self.hubbard_set.all())
        return self._hubbards

    @hubbards.setter
    def hubbards(self, hubbards):
        self._hubbards = hubbards

    # accessors/aggregators
    @property
    def comp(self):
        return self.output.comp

    @property
    def hub_comp(self):
        hcomp = defaultdict(int)
        for h in self.hubbards:
            if not h:
                continue
            for a in self.input:
                if ( h.element == a.element and
                        h.ox in [None, a.ox]):
                    hcomp[h] += 1
        return dict(hcomp.items())

    @property
    def unit_comp(self):
        return unit_comp(self.comp)

    @property
    def needs_hubbard(self):
        return any( h for h in self.hubbards )

    #= input files as strings =#
    @property
    def POSCAR(self):
        return write_poscar(self.input)

    @property
    def INCAR(self):
        return self.get_incar()

    @property
    def KPOINTS(self):
        return self.get_kpoints()

    @property
    def POTCAR(self):
        return self.get_potcar()

    # INCAR / settings
    @property
    def MAGMOMS(self):
        moments = [ a.magmom for a in self.input.atoms ]
        if all([ m in [0, None] for m in moments ]):
            return ''
        magmoms = [[ 1, moments[0]]]
        for n in range(1, len(moments)):
            if moments[n] == moments[n-1]:
                magmoms[-1][0] += 1
            else:
                magmoms.append([1, moments[n]])
        momstr = ' '.join('%i*%.4f' % (v[0],v[1]) for v in magmoms)
        return '  MAGMOM = %s' % momstr

    @property
    def phase(self):
        p = thermo.Phase(energy=self.delta_e,
                composition=parse_comp(self.composition_id),
                description=str(self.input.spacegroup),
                per_atom=True)
        p.id = self.id
        return p

    def get_incar(self):
        s = dict((k.lower(), v) for k, v in self.settings.items() if not k in
                ['gamma', 'kppra', 'scale_encut', 'potentials', 'hubbards'])

        incar = '#= General Settings =#\n'
        for key in ['prec', 'istart', 'icharg', 'lsorbit', 'nelect']:
            if key in s:
                incar += ' %s\n' % vasp_format(key, s.pop(key))

        if self.MAGMOMS and not 'ispin' in s:
            s['ispin'] = 2
        incar += ' ISPIN = %d\n' % s.pop('ispin', 1)
        if self.MAGMOMS:
            incar += self.MAGMOMS

        if  any(hub for hub in self.hubbards):
            incar += '\n#= LDA+U Fields =#\n'
            incar += ' LDAU = .TRUE.\n'
            incar += ' LDAUPRINT = 1\n'
            hubbards = sorted(self.hubbards, key=lambda x: x.element_id)
            uvals = ' '.join(str(hub.u) for hub in hubbards)
            jvals = ' '.join('0' for hub in hubbards)
            lvals = ' '.join(str(hub.l) for hub in hubbards)
            incar += ' LDAUU = %s\n' % uvals
            incar += ' LDAUJ = %s\n' % jvals
            incar += ' LDAUL = %s\n' % lvals

        incar += '\n#= Parallelization =#\n'
        for key in ['lplane', 'nsim', 'ncore', 'lscalu', 'npar']:
            if key in s:
                incar += ' %s\n' % vasp_format(key, s.pop(key))

        incar += '\n#= Ionic Relaxation Settings =#\n'
        for key in ['nsw', 'ibrion', 'isif', 'isym', 
                    'symprec', 'potim', 'ediffg']:
            if key in s:
                incar += ' %s\n' % vasp_format(key, s.pop(key))

        incar += '\n#= Electronic Relxation Settings =#\n'
        for key in ['encut', 'nelm', 'nelmin', 'lreal', 'ediff', 'algo']:
            if key in s:
                incar += ' %s\n' % vasp_format(key, s.pop(key))

        incar += '\n#= Write flags =#\n'
        for key in ['lcharg', 'lwave', 'lelf', 'lhvar', 'lvtot']:
            if key in s:
                incar += ' %s\n' % vasp_format(key, s.pop(key))

        incar += '\n#= DOS =#\n'
        for key in ['nbands', 'ismear', 'sigma']:
            if key in s:
                incar += ' %s\n' % vasp_format(key, s.pop(key))

        if s.get('ldipol', False):
            incar += '\n# dipole fields\n'
            incar += ' LDIPOL = .TRUE.\n'
            for k in ['idipol', 'espilon']:
                if k in s:
                    incar += ' %s\n' % vasp_format(k, s.pop(k))

        incar += '\n#= Uncategorized/OQMD codes  =#\n'
        for k, v in s.items():
            incar += ' %s\n' % (vasp_format(k, v))
        return incar

    @INCAR.setter
    def INCAR(self, incar):
        settings = {}
        custom = ''
        magmoms = []
        ldauus = []
        ldauls = []
        ldaujs = []
        for line in open(incar):
            line = line.lower()
            line = line.split('=')
            settings[line[0].strip()] = line[1].strip()

        if self.input is not None:
            atom_types = []
            for atom in self.input.atoms:
                if atom.element.symbol in atom_types:
                    continue
                atom_types.append(atom.element.symbol)
            if ldauus and ldauls:
                assert len(ldauls) == len(atom_types)
                assert len(ldauus) == len(atom_types)
                for elt, u, l in zip(atom_types, ldauus, ldauls):
                    hub, created = Hubbard.objects.get_or_create(element_id=elt,
                            u=u, l=l)
                    self.hubbards.append(hub)
            if magmoms:
                real_moms = []
                for mom in magmoms:
                    if '*' in mom:
                        num, mom = mom.split('*')
                        real_moms += [mom]*int(num)
                    else:
                        real_moms.append(mom)
                for atom, mom in zip(self.input.atoms, real_moms):
                    atom.magmom = float(mom)
                    if atom.id is not None:
                        atom.save()
        self.settings = settings

    def get_kpoints(self):
        n = float(len(self.input.atoms))
        recs = self.input.reciprocal_lattice
        rec_mags = [ norm(recs[0]), norm(recs[1]), norm(recs[2])]
        r0 = min(rec_mags)
        refr = np.array([ np.round(r/r0, 4) for r in rec_mags ])
        scale = 1.0
        kpts = np.ones(3)
        kppra = self.settings.get('kppra', 8000)
        while n*np.product(kpts) < kppra:
            prev_kpts = kpts.copy()
            refk = np.array(np.ones(3)*refr)*scale
            kpts = np.array(map(np.ceil, refk))
            scale += 1
        if kppra - np.product(prev_kpts)*n < np.product(kpts)*n - kppra:
            kpts = prev_kpts.copy()
        if self.settings.get('gamma', True):
            kpoints = 'KPOINTS \n0 \nGamma\n'
        else:
            kpoints = 'KPOINTS \n0 \nMonkhost-Pack\n'
        kpoints += ' '.join( str(int(k)) for k in kpts ) + '\n'
        kpoints += '0 0 0'
        return kpoints

    @KPOINTS.setter
    def KPOINTS(self, kpoints):
        raise NotImplementedError

    def get_potcar(self, distinct_by_ox=False):
        potstr = ''
        if not distinct_by_ox:
            elts = sorted(self.input.comp.keys())
        else:
            e_o_pairs = set([ (a.element_id, a.ox)
                for a in self.input ])
            elts = sorted([ p[0] for p in e_o_pairs ])

        for elt in elts:
            pot = [ p for p in self.potentials if p.element_id == elt ][0]
            potstr += pot.potcar
            potstr += ' End of Dataset\n'
        return potstr

    @POTCAR.setter
    def POTCAR(self, potcar):
        pots = Potential.read_potcar(potcar)
        for pot in pots:
            self.potentials.append(pot)

    @POSCAR.setter
    def POSCAR(self, poscar):
        self.input = read_poscar(poscar)

    #= Results Reading =#

    def read_vasprun_xml(self):
        raise NotImplementedError

    def get_outcar(self):
        '''
        Sets the calculations outcar attribute to a list of lines from the
        outcar. 

        Example:
        >>> calc = Calculation.read('calculation_path')
        >>> print calc.outcar
        None
        >>> calc.get_outcar()
        >>> len(calc.outcar)
        12345L
        
        '''
        if not self.outcar is None:
            return self.outcar
        if not exists(self.path):
            return
        elif exists(self.path+'/OUTCAR'):
            self.outcar = open(self.path+'/OUTCAR').read().split('\n')
        elif exists(self.path+'/OUTCAR.gz'):
            self.outcar = gzip.open(self.path + '/OUTCAR.gz', 'rb').read().split('\n')
        else:
            raise VaspError('No such file exists')

    def read_energies(self):
        '''
        Returns a numpy.ndarray of energies over all ionic steps.

        Example:
        >>> calc = Calculation.read('calculation_path')
        >>> calc.read_energies()
        array([-12.415236, -12.416596, -12.416927])
        
        '''        
        self.get_outcar()
        energies = []
        for line in self.outcar:
            if 'free  energy' in line:
                energies.append(ffloat(line.split()[4]))
        self.energies = np.array(energies)
        return self.energies

    def read_natoms(self):
        '''Reads the number of atoms, and assigns the value to natoms.'''
        self.get_outcar()
        for line in self.outcar:
            if 'NIONS' in line:
                self.natoms = int(line.split()[-1])
                return self.natoms

    def read_n_ionic(self):
        '''Reads the number of ionic steps, and assigns the value to nsteps.'''
        self.get_outcar()
        self.nsteps = len([ l for l in self.outcar if 'free  energy' in l ])
        return self.nsteps

    def read_input_structure(self):
        if os.path.exists(self.path+'/POSCAR'):
            self.input = read_poscar(self.path+'/POSCAR')
            self.input.entry = self.entry

    def read_elements(self):
        '''
        Reads the elements of the atoms in the structure. Returned as a list of
        atoms of shape (natoms,). 

        Example:
        >>> calc = Calculation.read('path_to_calculation')
        >>> calc.read_elements()
        ['Fe', 'Fe', 'O', 'O', 'O']
        '''
        self.get_outcar()
        elt_list = []
        elements = []
        for line in self.outcar:
            if 'POTCAR:' in line:
                elt = line.split()[2].split('_')[0]
                elt_list.append(elt)
            if 'ions per type' in line:
                elt_list = elt_list[:len(elt_list)/2]
                self.elements = [ Element.get(e) for e in elt_list ]
                counts = map(int, line.split()[4:])
                assert len(counts) == len(elt_list)
                for n, e in zip(counts, elt_list):
                    elements += [e]*n
                break
        self.elements = elements
        return self.elements

    def read_lattice_vectors(self):
        '''
        Reads and returns a numpy ndarray of lattice vectors for every ionic 
        step of the calculation.

        Example:
        >>> calc = Calculation.read('path_to_calculation')
        >>> calc.read_lattice_vectors()
        array([[[ 4.17458823,  0.        ,  0.        ],
                [-2.08729412,  3.61529939,  0.        ],
                [ 0.        ,  0.        ,  5.20720757]],

               [[ 4.17458823,  0.        ,  0.        ],
                [-2.08729412,  3.61529939,  0.        ],
                [ 0.        ,  0.        ,  5.20720757]],

               [[ 4.15950113,  0.        ,  0.        ],
                [-2.07975057,  3.60223358,  0.        ],
                [ 0.        ,  0.        ,  5.18670694]],

               [[ 4.16254664,  0.        ,  0.        ],
                [-2.08127332,  3.60487107,  0.        ],
                [ 0.        ,  0.        ,  5.18394359]]])

        '''
        self.get_outcar()
        lattice_vectors = []
        for i, line in enumerate(self.outcar):
            if 'direct lattice vectors' in line:
                tlv = []
                for n in range(3):
                    vector = self.outcar[i+n+1].split()[:3]
                    tlv.append(map(float, vector))
                lattice_vectors.append(tlv)
        return np.array(lattice_vectors)

    def read_charges(self):
        '''
        Reads and returns VASP's calculated charges for each atom. Returns the
        RAW charge, not NET charge.
        
        Example:
        >>> calc = Calculation.read('path_to_calculation')
        >>> calc.read_charges()

        '''
        self.get_outcar()
        self.read_natoms()
        self.read_n_ionic()
        if self.settings is None:
            self.read_outcar_settings()
        if not self.settings['lorbit'] == 11:
            return np.array([[0]*self.natoms]*self.nsteps)

        charges = []
        for n, line in enumerate(self.outcar):
            if 'total charge ' in line:
                chgs = []
                for i in range(self.natoms):
                    chgs.append(float(self.outcar[n+4+i].split()[-1]))
                charges.append(chgs)
        return np.array(charges[:-1])

    def read_magmoms(self):
        self.get_outcar()
        self.read_natoms()
        self.read_n_ionic()
        if self.settings is None:
            self.read_outcar_settings()
        if self.settings['ispin'] == 1 or self.settings['lorbit'] != 11:
            return np.array([[0]*self.natoms]*self.nsteps)

        magmoms = []
        for n, line in enumerate(self.outcar):
            if 'magnetization (x)' in line:
                mags = []
                for i in range(self.natoms):
                    mags.append(float(self.outcar[n+4+i].split()[-1]))
                magmoms.append(mags)
            if 'number of electron' in line:
                if 'magnetization' in line:
                    self.magmom = float(line.split()[-1])
        return magmoms[:-1]

    def read_forces(self):
        self.get_outcar()
        self.read_natoms()
        forces = []
        force_loop = [None]*self.natoms
        for line in self.outcar:
            if 'POSITION' in line:
                force_loop = []
            elif len(force_loop) < self.natoms:
                if '------' in line:
                    continue
                force_loop.append(map(float, line.split()[3:]))
                if len(force_loop) == self.natoms:
                    forces.append(force_loop)
        return np.array(forces)

    def read_positions(self):
        self.get_outcar()
        self.read_natoms()
        positions = []
        position_loop = [None]*self.natoms
        for line in self.outcar:
            if 'POSITION' in line:
                position_loop = []
            elif len(position_loop) < self.natoms:
                if '------' in line:
                    continue
                position_loop.append(map(float, line.split()[:3]))
                if len(position_loop) == self.natoms:
                    positions.append(position_loop)
        return np.array(positions)

    def read_stresses(self):
        self.get_outcar()
        stresses = []
        check = False
        for line in self.outcar:
            if 'FORCE on cell' in line:
                check = True
            if check and 'Total' in line:
                stresses.append(map(ffloat, line.split()[1:]))
                check = False
        return np.array(stresses)

    def read_kpoints(self):
        kpts = []
        weights = []
        found = False
        for i, line in enumerate(self.outcar):
            if 'irreducible k-points' in line:
                self.irreducible_kpoints = int(line.split()[1])
            if 'k-points in reciprocal lattice and weights' in line:
                for j in range(self.irreducible_kpoints):
                    x,y,z,w = map(float, self.outcar[i+j+1].split())
                    kpts.append([x,y,z])
                    weights.append(w)
                else:
                    break
        self.kpoints = kpts
        self.kpt_weights = weights

    def read_occupations(self):
        if self.kpoints is None:
            self.read_kpoints()
        if self.settings is None:
            self.read_outcar_settings()
        occs = []
        bands = []
        for i, line in enumerate(self.outcar):
            if 'k-point' in line:
                if not 'occupation' in self.outcar[i+1]:
                    continue
                if ' 1 ' in line:
                    occs = []
                    bands = []
                tocc = []
                tband = []
                for j in range(self.settings['nbands']):
                    b, e, o = map(ffloat, self.outcar[i+j+2].split())
                    tocc.append(o)
                    tband.append(e)
                occs.append(tocc)
                bands.append(tband)
        self.occupations = np.array(occs)
        self.bands = np.array(bands)

    def read_outcar_results(self):
        try:
            self.read_natoms()
            self.read_convergence()
            elts = self.read_elements()
            energies = self.read_energies()
            lattice_vectors = self.read_lattice_vectors()
            stresses = self.read_stresses()
            positions = self.read_positions()
            forces = self.read_forces()
            magmoms = self.read_magmoms()
            charges = self.read_charges()
        except:
            self.add_error('incomplete')

        if len(self.energies) > 0:
            self.energy = self.energies[-1]
            self.energy_pa = self.energy/self.natoms
        if not self.magmom is None:
            self.magmom_pa = self.magmom/self.natoms
        if self.nsteps > 1:
            output = Structure()
            output.total_energy = energies[-1]
            output.cell = lattice_vectors[-1]
            output.stresses = stresses[-1]
            inv = numpy.linalg.inv(output.cell).T
            atoms = []
            for coord, forces, charge, magmom, elt in zip(positions[-1], 
                                                          forces[-1],
                                                          charges[-1],
                                                          magmoms[-1],
                                                          elts):
                a = Atom(element_id=elt, charge=charge, magmom=magmom)
                a.coord = np.dot(inv, coord)
                a.forces = forces
                atoms.append(a)
            output.atoms = atoms
            self.output = output
        else:
            self.output = self.input

    def read_convergence(self):
        self.get_outcar()
        if not self.settings:
            self.read_outcar_settings()
        check_ionic = False
        if self.settings.get('nsw', 1) > 1:
            check_ionic = True

        v_init = None
        for line in self.outcar:
            if 'volume of cell' in line:
                v_init = float(line.split(':')[1].strip())
                break

        sc_converged, forces_converged = False, False
        v_fin = None
        for line in self.outcar[::-1]:
            if 'Iteration' in line:
                break
            if 'EDIFF is reached' in line:
                sc_converged = True
            if 'reached required accuracy' in line:
                forces_converged = True
            if 'volume of cell' in line:
                v_fin = float(line.split(':')[1].strip())

        if v_fin is None or v_init is None:
            basis_converged = False
        else:
            basis_converged = ( abs(v_fin - v_init)/v_init < 0.05 )

        if self.configuration in ['initialize', 
                                  'coarse_relax', 
                                  'fine_relax',
                                  'standard']:
            basis_converged = True

        if (sc_converged and 
                ((forces_converged and check_ionic) or not check_ionic) and
                basis_converged):
            self.converged = True
        else:
            self.add_error('convergence')
            self.converged = False

    def read_outcar_settings(self):
        self.get_outcar()
        settings = {'potentials':[]}
        elts = []
        for line in self.outcar:
            ### general options
            if 'PREC' in line:
                settings['prec'] = line.split()[2]
            elif 'ENCUT' in line:
                settings['encut'] = float(line.split()[2])
            elif 'ISTART' in line:
                settings['istart'] = int(line.split()[2])
            elif 'ISPIN' in line:
                settings['ispin'] = int(line.split()[2])

            # electronic relaxation 1
            elif 'NELM' in line:
                settings['nelm'] = int(line.split()[2].rstrip(';'))
                settings['nelmin'] = int(line.split()[4].rstrip(';'))
            elif 'LREAL  =' in line:
                lreal = line.split()[2]
                if lreal == 'F':
                    settings['lreal'] = False
                elif lreal == 'A':
                    settings['lreal'] = 'auto'
                elif lreal == 'T':
                    settings['lreal'] = True

            # ionic relaxation 
            elif 'EDIFF  =' in line:
                settings['ediff'] = float(line.split()[2])
            elif 'ISIF' in line:
                settings['isif'] = int(line.split()[2])
            elif 'IBRION' in line:
                settings['ibrion'] = int(line.split()[2])
            elif 'NSW' in line:
                settings['nsw'] = int(line.split()[2].rstrip(';'))
            elif 'PSTRESS' in line:
                settings['pstress'] = float(line.split()[1])
            elif 'POTIM' in line:
                settings['potim'] = float(line.split()[2])

            # DOS Flags
            elif 'ISMEAR' in line:
                line = line.split()
                settings['ismear'] = int(line[2].rstrip(';'))
                settings['sigma'] = float(line[5])
            elif 'NBANDS=' in line:
                if not 'INCAR' in line:
                    settings['nbands'] = int(line.split()[-1])

            # write flags
            elif 'LCHARG' in line:
                settings['lcharg'] = ( line.split()[2] != 'F' )
            elif 'LWAVE' in line:
                settings['lwave'] = ( line.split()[2] == 'T' )
            elif 'LVTOT' in line:
                settings['lvtot'] = ( line.split()[2] == 'T' )
            elif 'LORBIT' in line:
                settings['lorbit'] = int(line.split()[2])

            # electronic relaxation 2
            elif 'ALGO' in line:
                algo_dict = {38:'normal',
                             68:'fast',
                             48:'very_fast',
                             58:'all',
                             53:'default'}
                settings['algo'] = algo_dict[int(line.split()[2])]

            # dipole flags
            elif 'LDIPOL' in line:
                settings['ldipol'] = ( line.split()[2] == 'T')
            elif 'IDIPOL' in line:
                settings['idipol'] = int(line.split()[2])
            elif ' EPSILON=' in line:
                settings['epsilon'] = float(line.split()[1])
            
            # potentials
            elif 'POTCAR:' in line:
                this_pot = {'name':line.split()[2]}
            elif 'Description' in line:
                settings['potentials'].append(this_pot)
            elif 'LEXCH' in line:
                key = line.split()[2]
                if key == '91':
                    this_pot['xc'] = 'GGA'
                elif key == 'CA':
                    this_pot['xc'] = 'LDA'
                elif key == 'PE':
                    this_pot['xc'] = 'PBE'
            elif 'LULTRA' in line:
                key = line.split()[2]
                this_pot['us'] = ( key == 'T' )
            elif 'LPAW' in line:
                key = line.split()[2]
                this_pot['paw'] = ( key == 'T' )
            # hubbards
            elif 'LDAUL' in line:
                settings['ldau'] = True
                settings['ldauls'] = [ int(v) for v in line.split()[7:] ]
            elif 'LDAUU' in line:
                settings['ldauus'] = [ float(v) for v in line.split()[7:] ]
            elif 'energy-cutoff' in line:
                break

        # assign potentials
        xcs = list(set([ p['xc'] for p in settings['potentials']]))
        uss = list(set([ p['us'] for p in settings['potentials']]))
        paws = list(set([ p['paw'] for p in settings['potentials']]))
        pot_names = [ p['name'] for p in settings['potentials']]
        elts = [ p['name'].split('_')[0] for p in settings['potentials'] ]
        if any([ len(s) > 1 for s in [xcs, uss, paws]]):
            raise VaspError('Not all potentials are of the same type')
        self.potentials = Potential.objects.filter(us=uss[0], 
                                              xc=xcs[0], 
                                              paw=paws[0],
                                              name__in=pot_names)

        # assign hubbards
        self.hubbards = []
        if 'ldauls' in settings:
            for elt, l, u in zip(elts, 
                                 settings['ldauls'], 
                                 settings['ldauus']):
                self.hubbards.append(Hubbard.get(elt, u=u, l=l))
        self.settings = settings

    def read_stdout(self, filename='stdout.txt'):
        if not os.path.exists('%s/%s' % (self.path, filename)):
            return []
        stdout = open('%s/%s' % (self.path, filename)).read()
        errors = []
        if 'Error reading item' in stdout:
            errors.append('input_error')
        if 'ZPOTRF' in stdout:
            errors.append('zpotrf')
        if 'SGRCON' in stdout:
            errors.append('sgrcon')
        if 'INVGRP' in stdout:
            errors.append('invgrp')
        if 'BRIONS problems: POTIM should be increased' in stdout:
            errors.append('brions')
        if 'FEXCF' in stdout:
            errors.append('fexcf')
        if 'FEXCP' in stdout:
            errors.append('fexcp')
        if 'PRICEL' in stdout:
            errors.append('pricel')
        if 'EDDDAV' in stdout:
            errors.append('edddav')
        if 'Sub-Space-Matrix is not hermitian in DAV' in stdout:
            errors.append('hermitian')
        if 'IBZKPT' in stdout:
            errors.append('IBZKPT error')
        if 'BRMIX: very serious problems' in stdout:
            errors.append('brmix')
        return errors

    def read_outcar(self):
        if self.input is None:
            self.read_input_structure()
        self.get_outcar()
        self.read_outcar_settings()
        self.read_outcar_results()
        self.read_n_ionic()

    def read_doscar(self):
        if os.path.getsize(self.path+'/DOSCAR') < 300:
            return
        self.dos = dos.DOS.read(self.path+'/DOSCAR')
        self.band_gap = self.dos.find_gap()
        return self.dos

    def clear_outputs(self):
        if not os.path.exists(self.path):
            return
        for file in os.listdir(self.path):
            if os.path.isdir(self.path+'/'+file):
                continue
            if file in ['INCAR', 'POSCAR', 'KPOINTS', 'POTCAR']:
                continue
            os.unlink('%s/%s' % (self.path, file))

    def clear_results(self):
        self.energy = None
        self.energy_pa = None
        self.magmom = None
        self.magmom_pa = None
        self.output = None
        self.dos = None
        self.band_gap = None
        self.irreducible_kpoints = None
        self.runtime = None
        self.nsteps = None
        self.converged = None
        self.delta_e = None
        self.stability = None

    @staticmethod
    def read(path):
        '''
        Reads the outcar specified by the objects path. Populates input field
        values, as well as outputs, in addition to finding errors and
        confirming convergence.

        Example:
        >>> calc = Calculation.read('/path/to/calculation')
        '''
        path = os.path.abspath(path)
        existing = Calculation.objects.filter(path=path)
        if existing.count() > 1:
            return existing
        elif existing.count() == 1:
            return existing[0]

        calc = Calculation(path=path)
        if calc.input is None:
            calc.read_input_structure()
        calc.label = os.path.basename(calc.path)
        calc.read_outcar()
        calc.read_stdout()
        if calc.converged:
            calc.read_doscar()
        #except Exception, err:
        #    calc.errors.append('exception: %s' % err)
        return calc

    def address_errors(self):
        errors = self.errors
        if not errors:
            logger.info('Found no errors')
            return self
        
        if self.label == '':
            self.label = os.path.basename(self.path)
        new_calc = self.copy()
        self.label = self.label + '_%d' % self.attempt
        new_calc.attempt += 1

        for err in errors:
            if err in ['duplicate','partial']:
                continue
            elif err == 'convergence':
                if self.nsteps > 1:
                    new_calc.remove_error('convergence')
                    new_calc.input = self.output
                    new_calc.input.label = self.label
            elif err == 'electronic_convergence':
                new_calc.fix_electronic_convergence()
            elif err == 'doscar_exc':
                new_calc.fix_bands()
            elif err == 'bands':
                new_calc.fix_bands()
            elif err == 'edddav':
                new_calc.fix_dav()
            elif err == 'errrmm':
                new_calc.fix_rmm()
            elif err == 'sgrcon':
                new_calc.fix_sgrcon()
            elif err == 'zpotrf':
                new_calc.fix_zpotrf()
            elif err == 'invgrp':
                new_calc.fix_invgrp()
            elif err == 'exception':
                new_calc.clean_start()
            elif err == 'fexcf':
                new_calc.fix_fexcf()
            elif err == 'fexcp':
                new_calc.fix_fexcp()
            elif err == 'brions':
                new_calc.fix_brions()
            elif err == 'brmix':
                new_calc.fix_brmix()
            elif err in ['pricel', 'invgrp', 'sgrcon']:
                new_calc.increase_symprec()
            elif err == 'hermitian':
                new_calc.fix_hermitian()
            else:
                print err

        new_calc.set_magmoms(self.settings.get('magnetism', 'ferro'))
        return new_calc

    #= File handling =#

    def compress(self):
        for file in os.listdir(self.path):
            if file in ['OUTCAR', 'CHGCAR', 'CHG', 'PROCAR', 'LOCPOT', 'ELFCAR']:
                os.system('gzip -f %s' % self.path+'/'+file)

    def copy(self):
        calc = copy.deepcopy(self)
        calc.id = None
        calc.label = ''
        if self.input is not None:
            calc.input = self.input
        if self.output is not None:
            calc.output = self.output
        calc.potentials = self.potentials
        calc.hubbards = self.hubbards
        return calc

    def move(self, path):
        path = os.path.abspath(path)
        os.system('mkdir %s &> /dev/null' % path)
        os.system('cp %s/* %s &> /dev/null' % (self.path, path))
        os.system('rm %s/* &> /dev/null' % self.path)
        self.path = path
        self.save()

    def backup(self):
        new_dir = '%s_' % self.attempt
        new_dir += '_'.join(self.errors)
        logger.info('backing up %s to %s' % (self.path, new_dir))
        self.move(self.path+'/'+new_dir)

    def clean_start(self):
        depth = self.path.count('/') - self.path.count('..')
        if depth < 6:
            raise ValueError('Too short path supplied to clean_start: %s' % self.path)
        else:
            os.system('rm -rf %s &> /dev/null' % self.path)

    #= Error correcting =#

    def fix_zhegev(self):
        raise NotImplementedError

    def fix_brmix(self):
        self.settings.update({'symprec':1e-7,
                              'algo':'normal'})
        self.remove_error('brmix')

    def fix_fexcf(self):
        self.settings.update({'potim':0.1,
                              'algo':'normal'})
        self.remove_error('fexcf')

    def fix_electronic_convergence(self):
        if not self.settings.get('algo') == 'normal':
            self.settings['algo'] = 'normal'
            self.remove_error('electronic_convergence')

    def increase_symprec(self):
        self.settings['symprec'] = 1e-7
        self.remove_error('invgrp')
        self.remove_error('pricel')
        self.remove_error('sgrcon')

    def fix_brions(self):
        self.settings['potim'] *= 2
        self.remove_error('brions')

    def fix_zpotrf(self):
        self.settings.update({'algo':'normal',
                              'potim':0.1})
        self.remove_error('zpotrf')

    def fix_bands(self):
        self.settings['nbands'] *= int(np.ceil(1.15))
        self.errors = []

    def fix_dav(self):
        if self.settings['algo'] == 'fast':
            self.settings['algo'] = 'very_fast'
        elif self.settings['algo'] == 'normal':
            self.settings['algo'] = 'very_fast'
        else:
            return
        self.remove_error('edddav')
        self.remove_error('electronic_convergence')

    def fix_rmm(self):
        if self.settings['algo'] == 'fast':
            self.settings['algo'] = 'normal'
        elif self.settings['algo'] == 'very_fast':
            self.settings['algo'] = 'normal'
        else:
            return
        self.remove_error('errrmm')
        self.remove_error('electronic_convergence')

    def fix_hermitian(self):
        if self.settings['algo'] == 'very_fast':
            return
        self.settings['algo'] = 'very_fast'
        self.remove_error('hermitian')
        self.remove_error('electronic_convergence')

    #### calculation management

    def write(self):
        os.system('mkdir %s &> /dev/null' % self.path)
        poscar = open(self.path+'/POSCAR','w')
        potcar = open(self.path+'/POTCAR','w')
        incar = open(self.path+'/INCAR','w')
        kpoints = open(self.path+'/KPOINTS','w')
        poscar.write(self.POSCAR)
        potcar.write(self.POTCAR)
        incar.write(self.INCAR)
        kpoints.write(self.KPOINTS)
        poscar.close()
        potcar.close()
        incar.close()
        kpoints.close()

    @property
    def estimate(self):
        return 24*8*3600

    @property
    def instructions(self):
        if self.converged:
            return {}
        elif self.errors:
            return {}

        instruction = {
                'path':self.path,
                'walltime':self.estimate,
                'header':'\n'.join(['gunzip -f CHGCAR.gz &> /dev/null',
                    'date +%s',
                    'ulimit -s unlimited']),
                'mpi':'mpirun -machinefile $PBS_NODEFILE -np $NPROCS',
                'binary':'vasp_53', 
                'pipes':' > stdout.txt 2> stderr.txt',
                'footer':'\n'.join(['gzip -f CHGCAR OUTCAR PROCAR ELFCAR',
                    'rm -f WAVECAR CHG',
                    'date +%s'])}

        if self.input.natoms < 10:
            instruction.update({'mpi':'','binary':'vasp_53_serial',
                'serial':True})
        return instruction

    def set_hubbards(self, convention='wang'):
        hubs = HUBBARDS.get(convention, {})
        elts = set( k[0] for k in hubs.keys() )
        ligs = set( k[1] for k in hubs.keys() )
        lig_int = ligs & set(self.input.comp.keys())

        if not lig_int:
            return
        elif len(lig_int) > 1:
            raise Exception('More than 1 ligand matches. No convention\
            established for this case!')
        
        if not elts & set(self.input.comp.keys()):
            return

        for atom in self.input:
            for hub in hubs:
                if ( atom.element_id == hub[0] and 
                        hub[2] in [ None, atom.ox ]):
                    self.hubbards.append(Hubbard.get(
                        hub[0], lig=hub[1], ox=hub[2],
                        u=hubs[hub]['U'], l=hubs[hub]['L']))
                    break
            else:
                self.hubbards.append(Hubbard.get(atom.element_id))
        self.hubbards = list(set(self.hubbards))

    def set_potentials(self, choice='vasp_rec', distinct_by_ox=False):
        if isinstance(choice, list):
            if len(self.potentials) == len(choice):
                return
        pot_set = POTENTIALS[choice]
        potentials = Potential.objects.filter(xc=pot_set['xc'],
                                              gw=pot_set['gw'],
                                              us=pot_set['us'],
                                              paw=pot_set['paw'])

        pnames = [ pot_set['elements'][e.symbol] for e in self.elements ]
        self.potentials = list(potentials.filter(name__in=pnames))

    def set_magmoms(self, ordering='ferro'):
        self.input.set_magnetism(ordering)
        if any(self.input.magmoms):
            self.ispin = 2

    def compute_formation(self, reference='standard'):
        #if not self.label == 'standard':
        #    print 'WARNING: This formation energy is with respect to\
        #            innappropriate reference states and cannot be trusted!'
        if not self.converged:
            return None
        hub_mus = chem_pots[reference]['hubbards']
        elt_mus = chem_pots[reference]['elements']
        adjust = 0
        adjust -= sum([ hub_mus.get(k.key, 0)*v for k,v in self.hub_comp.items() ])
        adjust -= sum([ elt_mus[k]*v for k,v in self.comp.items() ])
        self.delta_e = ( self.energy + adjust ) / self.natoms
        return self.delta_e

    @staticmethod
    def setup(input=None, configuration='standard', path=None, entry=None,
            hubbard='wang', potentials='vasp_rec', settings={}, **kwargs):
        
        if path is None:
            path = os.path.abspath(entry.path)
        else:
            path = os.path.abspath(path)

        if Calculation.objects.filter(path=path).exists():
            calc = Calculation.objects.get(path=path)
        else:
            if not os.path.exists(path):
                os.mkdir(path)
            calc = Calculation()
            calc.configuration = configuration
            calc.path = path
            calc.input = input
            calc.kwargs = kwargs
            if entry:
                calc.entry = entry

        if configuration not in VASP_SETTINGS:
            raise NotImplementedError('That configuration does not exist!')

        settings = {}
        if calc.input.natoms > 20:
            settings['lreal'] = 'auto'
        settings.update(VASP_SETTINGS[configuration])
        settings.update(settings)
        calc.set_potentials(calc.settings.get('potentials', 'vasp_rec'))
        calc.set_hubbards(calc.settings.get('hubbards', 'wang'))
        calc.set_magmoms(calc.settings.get('magnetism', 'ferro'))

        if 'scale_encut' in settings:
            enmax = max(pot.enmax for pot in calc.potentials)
            calc.encut = int(settings.get('scale_encut')*enmax)

        calc.settings = settings

        #= Error handling stuff =#
        if set(calc.errors) <= set(['attempts', 'ionic_convergence']):
            calc.remove_error('attempts')
            calc.remove_error('ionic_convergence')

        # has the calculation been run yet?
        try:
            calc.get_outcar()
        except VaspError:
            if not ( os.path.exists(calc.path+'/INCAR') and
                    os.path.exists(calc.path+'/POTCAR') and 
                    os.path.exists(calc.path+'/KPOINTS') and
                    os.path.exists(calc.path+'/POSCAR')):
                calc.write()
            return calc

        calc.read_outcar()
        calc.read_doscar()
        calc.read_stdout()

        if calc.converged:
            return calc
        elif not calc.errors:
            calc.write()
            return calc
        
        fixed_calc = calc.address_errors()
        fixed_calc.clear_results()
        calc.backup()
        fixed_calc.clear_outputs()
        fixed_calc.write()
        return fixed_calc



import os, os.path
import sys
import site

INSTALL_DIR = os.path.abspath(__file__).replace('/qmpy/db/wsgi.py', '')

os.environ['HOME'] = '/dev/shm'
import matplotlib
matplotlib.use('Agg')
import matplotlib.cbook

# # Add the app's directory to the PYTHONPATH
sys.path.append(INSTALL_DIR+'/qmpy/db')
sys.path.append(INSTALL_DIR)

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "qmpy.db.settings")

# This application object is used by the development server
# as well as any WSGI server configured to use this file.
#from django.core.wsgi import get_wsgi_application
#application = get_wsgi_application()
import django.core.handlers.wsgi

application = django.core.handlers.wsgi.WSGIHandler()



import yaml
import os.path

location = os.path.dirname(__file__)

data = open(location+'/elements/groups.yml').read()
element_groups = yaml.load(data)

data = open(location+'/elements/data.yml').read()
elements = yaml.load(data)

data = open(location+'/elements/chemical_potentials.yml').read()
chem_pots = yaml.load(data)



from datetime import datetime, timedelta
import random
import subprocess
import os
import os.path
import time

from django.db import models
from django.contrib.auth.models import User

from qmpy.db.custom import DictField

class AllocationError(Exception):
    '''Problem with the allocation'''

class SubmissionError(Exception):
    '''Failed to submit a job'''

class Host(models.Model):
    '''
    Host model - stores all host information for a cluster.

    Stored attributes:
        name
        ip_address
        hostname
        binaries    :   dict of binary names and paths 
        ppn         :   # of processors per node (int)
        nodes
        walltime
        sub_script
        sub_text
        check_queue
        checked_time
        running
        utilization
    '''
    name = models.CharField(max_length=20, db_index=True)
    ip_address = models.CharField(max_length=120)
    hostname = models.CharField(max_length=120)
    binaries = DictField()
    ppn = models.IntegerField(default=8)
    nodes = models.IntegerField(default=30)
    walltime = models.IntegerField(default=3600*24)
    sub_script = models.CharField(max_length=120)
    sub_text = models.TextField(default='/usr/local/bin/qsub')
    check_queue = models.CharField(max_length=180,
            default='/usr/local/maui/bin/showq')
    checked_time = models.DateTimeField(default=datetime.min)
    running = DictField(blank=True, null=True)
    utilization = models.IntegerField(default=0)
    state = models.IntegerField(default=1)

    class Meta:
        app_label = 'qmpy'
        db_table = 'hosts'

    def __str__(self):
        return self.name

    @classmethod
    def interactive_create(cls):
        '''
        Classmethod to create a Host model. Script will ask you questions about
        the host to add, and will return the created Host.

        '''
        host = {}
        host['name'] = raw_input('Hostname:')
        if cls.objects.filter(name=host['name']).exists():
            print 'Host by that name already exists!'
            exit(-1)
        host['ip_address'] = raw_input('IP Address:')
        if cls.objects.filter(ip_address=host['ip_address']).exists():
            print 'Host at that address already exists!'
            exit(-1)
        host['ppn'] = raw_input('Processors per node:')
        host['nodes'] = raw_input('Max nodes to run on:')
        host['sub_script'] = raw_inputs('Command to submit a script '+
                '(e.g. /usr/local/bin/qsub):')
        host['check_queue'] = raw_input('Command for showq (e.g.'+
                '/usr/local/maui/bin/showq):')
        host['sub_text'] = raw_input('Path to qfile template:')
        h = cls(**host)
        h.save()

    @property
    def accounts(self):
        return list(self.account_set.all())

    @property
    def jobs(self):
        jobs = []
        for acct in self.accounts:
            jobs += list(acct.job_set.filter(state=1))
        return jobs

    @property
    def active(self):
        if self.state < 1:
            return False
        elif self.utilization > self.nodes*self.ppn:
            return False
        else:
            return True

    @property
    def percent_utilization(self):
        return 100. * float(self.utilization) / (self.nodes*self.ppn)

    @property
    def utilization(self):
        util = 0
        for acct in self.account_set.all():
            for job in acct.job_set.filter(state=1):
                util += job.ncpus
        return util

    @property
    def qfile(self):
        return open(self.sub_text).read()

    def get_binary(self, key):
        return self.binaries[key]

    def check_host(self):
        '''Pings the host to see if it is online. Returns False if it is
        offline.'''
        ret = subprocess.call("ping -c 1 -w 1 %s" % self.ip_address,
                shell=True,
                stdout=open('/dev/null', 'w'),
                stderr=subprocess.STDOUT)
        if ret == 0:
            return True
        else:
            self.state = -2
            self.save()
            return False

    @property
    def running_now(self):
        if datetime.now() + timedelta(seconds=-30) > self.checked_time:
            self.check_running()
        return self.running

    def check_running(self, verbosity=0):
        '''
        Uses the hosts data and one of the associated accounts to check the PBS
        queue on the Host. If it has been checked in the last 2 minutes, it
        will return the previously returned result.

        '''
        self.checked_time = datetime.now()
        if self.state < 0:
            self.running = {}
            self.save()
            return
        account = random.choice(self.accounts)
        raw_data = account.execute(self.check_queue, verbosity=verbosity)
        running = {}
        if not raw_data:
            return
        for line in raw_data.split('\n'):
            if 'Active Jobs' in line:
                continue
            line = line.split()
            if len(line) != 9:
                continue
            try:
                running[int(line[0])] = {
                        'user':line[1],
                        'state':line[2],
                        'proc':int(line[3])}
            except:
                pass
        self.running = running
        self.save()
        
    def get_running(self):
        if self.running is not None:
            return self.running
        else:
            return {}

#===============================================================================#

class Account(models.Model):
    user = models.ForeignKey(User)
    host = models.ForeignKey(Host)
    username = models.CharField(max_length=20)
    run_path = models.CharField(max_length=120)

    state = models.IntegerField(default=1)
    class Meta:
        app_label = 'qmpy'
        db_table = 'accounts'

    def __str__(self):
        return '{user}@{host}'.format(user=self.user.username, 
                host=self.host.name)

    @property
    def active(self):
        if self.state < 1:
            return False
        elif not self.host.active:
            return False
        else:
            return True

    def submit(self, path=None, run_path=None, 
            qfile=None, verbosity=0):
        self.execute('mkdir %s' % run_path, verbosity=verbosity,
                ignore_output=True)
        self.copy(folder=path, file='*',
                destination=run_path,
                verbosity=verbosity)
        cmd = 'cd {path} && {sub} {qfile}'.format(
                path=run_path, 
                sub=self.host.sub_script,
                qfile=qfile)
        stdout = self.execute(cmd, verbosity=verbosity)
        jid = int(stdout.split()[0].split('.')[0])
        if self.host.name == 'quest':
            time.sleep(60)
        return jid

    def execute(self, command='exit 0', verbosity=0, ignore_output=False):
        ssh = 'ssh {user}@{host} "{cmd}"'.format(
                user=self.username,
                host=self.host.ip_address,
                cmd=command)

        if verbosity > 0:
            print ssh

        call = subprocess.Popen(ssh, shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        stdout,stderr = call.communicate()

        if verbosity > 0:
            print 'stdout:', stdout
            print 'stderr:', stderr
        if stderr and not ignore_output:
            print 'WARNING:'
            print stderr
        return stdout

    def copy(self, destination=None, to=None, # where to send the stuff
            fr=None, file=None, folder=None, # what to send
            clear_dest_dir=False, verbosity=0, move=False): # some conditions on sending it

        if destination is None:
            destination = self.run_path
        if to is None:
            to = self
        if fr is None:
            if to == 'local':
                fr = self
            else:
                fr = 'local'

        assert (isinstance(to, Account) or to == 'local')
        assert (isinstance(fr, Account) or fr == 'local')

        send_dir = False
        if file is None and folder is None:
            print 'Must specify something to copy'
        elif file is None:
            send_dir = True
        elif folder is None:
            folder = os.path.dirname(file)
            file = os.path.basename(file)
        
        if clear_dest_dir:
            if to == 'local':
                command = subprocess.Popen('rm -f %s/*' % destination,
                                                  stderr=subprocess.PIPE,
                                                  stdout=subprocess.PIPE)
                stdout, stderr = command.communicate()
            else:
                stdout, stderr = self.execute('rm -f %/*' % destination)
            if verbosity:
                print 'output:', stdout
            
        if fr == 'local':
            scp = 'scp '
        else:
            scp = 'scp {user}@{host}:'.format(
                    user=fr.username, host=fr.host.ip_address)

        if not file:
            scp += '-r '

        if send_dir:
            scp += os.path.abspath(folder)
        else:
            scp += '{path}/{file}'.format(
                    path=os.path.abspath(folder), file=file)

        if to == 'local':
            scp += ' '+destination
        else:
            scp += ' {user}@{host}:{path}'.format(
                user=to.username, host=to.host.ip_address, 
                path=os.path.abspath(destination))

        if verbosity > 0:
            print 'copy command:', scp
        cmd = subprocess.Popen(scp,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        stdout, stderr = cmd.communicate()
        if verbosity:
            print 'stdout:', stdout
            print 'stderr:', stderr

        if move:
            if send_dir:
                rmcmd = 'rm -rf {path}'.format(path=os.path.abspath(folder))
            else:
                rmcmd = 'rm -f {path}/{file}'.format(file=file,
                        path=os.path.abspath(folder))
            if verbosity > 0:
                print 'wiping source: ', rmcmd
            stdout = fr.execute(rmcmd)
            if verbosity > 0:
                print 'output:', stdout

#===============================================================================#

class Allocation(models.Model):
    name = models.CharField(max_length=50, db_index=True)
    key = models.CharField(max_length=100, default='', blank=True, null=True)

    host = models.ForeignKey(Host, null=True)
    users = models.ManyToManyField(User, null=True)
    state = models.IntegerField(default=1)
    
    class Meta:
        app_label = 'qmpy'
        db_table = 'allocations'

    def __str__(self):
        return self.name
    
    @classmethod
    def create_interactive(cls):
        name = raw_input('Name your allocation:')
        if cls.objects.filter(name=name).exists():
            print 'Allocation by that name already exists!'
            exit(-1)
        host = raw_input('Which cluster is this allocation on?')
        if not Host.objects.filter(name=host).exists():
            print "This host doesn't exist!"
            exit(-1)
        host = Host.objects.get(name=host)
        alloc = cls(name=name, host=host)
        alloc.save()
        print 'Now we will assign users to this allocation'
        for acct in Account.objects.filter(host=host):
            inc = raw_input('Can %s use this allocation? y/n [y]:' % 
                    acct.user.name )
            if inc == 'y' or inc == '':
                alloc.users.add(acct.user)
        print 'If this allocation requires a special password, enter',
        key = raw_input('it now:')
        alloc.key=key
        alloc.save()

    @property
    def active(self):
        if self.state < 1:
            return False
        elif not self.host.active:
            return False
        else:
            return True

    def get_user(self):
        return random.choice(self.users.filter(state=1))

    def get_account(self, users=None):
        if users is None:
            users = self.users.filter(state=1).all()
        user = random.choice(list(users))
        return user.account_set.get(host=self.host)

    @property
    def percent_utilization(self):
        return self.host.percent_utilization

#===============================================================================#

class Project(models.Model):
    name = models.CharField(max_length=60)
    priority = models.IntegerField(default=0)

    users = models.ManyToManyField(User)
    allocations = models.ManyToManyField(Allocation)
    state = models.IntegerField(default=1)
    class Meta:
        app_label = 'qmpy'
        db_table = 'projects'

    def __str__(self):
        return self.name

    @classmethod
    def get(cls, name):
        if isinstance(name, cls):
            return name
        proj, new = cls.objects.get_or_create(name=name)
        if new:
            proj.save()
        return proj

    @property
    def completed(self):
        return self.task_set.filter(state=2)

    @property
    def running(self):
        from queue import Job
        return Job.objects.filter(state=1, task__project_set=self)

    @classmethod
    def interactive_create(cls):
        name = raw_input('Name your project: ')
        if cls.objects.filter(name=name).exists():
            print 'Project by that name already exists!'
            exit(-1)
        proj = cls(name=name)
        proj.save()
        proj.priority = raw_input('Project priority (1-100): ')
        users = raw_input('List project users (e.g. sjk648 jsaal531 bwm291): ')
        for u in users.split():
            if not User.objects.filter(name=u).exists():
                print 'User named', u, 'doesn\'t exist!'
            else:
                proj.users.add(User.objects.get(name=u))

        alloc = raw_input('List project allocations (e.g. byrd victoria b1004): ')
        for a in alloc.split():
            if not Allocation.objects.filter(name=a).exists():
                print 'Allocation named', a, 'doesn\'t exist!'
            else:
                proj.allocations.add(Allocation.objects.get(name=a))

    @property
    def active(self):
        if self.state < 0:
            return False
        elif not any( a.active for a in self.allocations.all() ):
            self.state = 0
            self.save()
            return False
        else:
            if self.state != 1:
                self.state = 1
                self.save()
            return True

    def get_allocation(self):
        available = [ a for a in self.allocations.all() if a.active ]
        if available:
            return random.choice(available)
        else:
            return []



#!/usr/env/bin python

import itertools
import numpy as np
from qmpy.utils.math import *

def compute_d(cell, hkl):
    '''
    Takes a 1x6 array of lattice parameters, or a 3x3 array of lattice vectors,
    in combination with a 3x1 array of h, k, l indices and computes the
    interplanar distance of for that hkl vector.

    Example:
    >>> compute_d([3,4,5, 90, 90, 90], [1, 2, 3])

    '''
    cell = np.array(cell)
    if cell.shape == (6,):
        a, b, c, al, be, ga = cell
    elif cell.shape == (3,3):
        a, b, c, al, be, ga = basis_to_latparams(cell)
    a, b, c, al, be, ga = map(float, [a,b,c,al,be,ga])
    h, k, l = hkl
    d = 0
    d += h**2/(a**2*np.sin(al)**2)
    d += 2*k*l*(np.cos(be)*np.cos(ga)-np.cos(al))/(b*c)
    d += k**2/(b**2*np.sin(be)**2)
    d += 2*h*l*(np.cos(al)*np.cos(ga)-np.cos(be))/(a*c)
    d += l**2/(c**2*np.sin(ga)**2)
    d += 2*h*k*(np.cos(al)*np.cos(be)-np.cos(ga))/(a*b)
    d /= (1-np.cos(al)**2 - np.cos(be)**2 - np.cos(ga)**2 +
             2*np.cos(al)*np.cos(be)*np.cos(ga))
    return (1/d)**0.5

def compute_2theta(cell, hkl, wavelength=1.5418):
    '''
    Compute 2*theta for a given cell, in an hkl direction, for a given
    wavelength.
    '''
    d = compute_d(cell, hkl)
    return np.arcsin(wavelength/(2*d))*180/np.pi

def get_all_peaks(cell, min_2t=10, max_2t=70, wavelength=1.5418):
    '''
    For the given cell, it loops over hkl vectors within a range of 2theta 
    values, for a given wavelength.
    '''
    cell = np.array(cell)
    max_d = wavelength/(2*np.sin(max_2t*np.pi/360))
    if cell.shape == (3,3):
        cell = basis_to_latparams(cell)
    im, jm, km = map(lambda x: int(np.ceil(max_d*x)), cell[:3])
    peaks = []
    for h in range(-im, im+1):
        for k in range(-jm, jm+1):
            for l in range(-km, km+1):
                if [h,k,l] == [0,0,0]:
                    continue
                peak = compute_2theta(cell, [h,k,l])
                if peak > max_2t or peak < min_2t:
                    continue
                if not np.isnan(peak):
                    peaks.append(peak)
    return np.array(peaks)



#!/usr/bin/env python


from django.db import models
import json
import os.path
import time
from datetime import datetime, timedelta
import random

from resources import Project, Account, Allocation
import qmpy

class TaskError(Exception):
    '''A project was needed but not provided'''

class Task(models.Model):
    '''
    Model for a Task to be done. A Task consists of a module, which is the name
    of a computing script. 

    The Task.state attribute indicates:
        -2  :   error
        -1  :   held
        0   :   ready to advance
        1   :   running
        2   :   done

    Database Attributes:
        module      : name of a computing script - must correspond to a
                      qmpy.computing.script function. 
        args        : a json format string of a sequence of arguments to a
                      function.
        state       : int
        priority    : The priority of a task (float)

        entry       : Entry object
        project_set : QuerySet of Projects

        created     : DateTime field populated when the task is created.
        finished    : DateTeim field populated when the task is completed.
    '''
    module = models.CharField(max_length=60)
    args = models.CharField(max_length=100, blank=True)
    state = models.IntegerField(default=0)
    priority = models.IntegerField(default=50)
    created = models.DateTimeField(blank=True, auto_now_add=True)
    finished = models.DateTimeField(blank=True, null=True)

    entry = models.ForeignKey('Entry')
    project_set = models.ManyToManyField(Project, null=True)

    _projects = None

    class Meta:
        app_label = 'qmpy'
        db_table = 'tasks'

    def save(self, *args, **kwargs):
        super(Task, self).save(*args, **kwargs)
        self.project_set = [ Project.get(p) for p in self.projects ]

    @property
    def projects(self):
        '''List of related projects.'''
        if self._projects is None:
            self._projects = list(self.project_set.all())
        return self._projects

    @projects.setter
    def projects(self, projects):
        self._projects = projects

    @property
    def eligible_to_run(self):
        if self.state != 0:
            return False
        if self.entry.holds:
            return False
        if not any( p.active for p in self.project_set.all()):
            return False
        return True

    def get_project(self):
        projects = [ p for p in self.projects if p.active ]
        return random.choice(projects)

    @staticmethod
    def create(entry, module='standard', args=[],
            priority=None, projects=None): 
        if projects is None:
            projects = entry.projects
        elif isinstance(projects, basestring):
            projects = Project.objects.get(name=projects)
        if priority is None:
            priority = len(entry.input)
        task, created = Task.objects.get_or_create(entry=entry, 
                args=json.dumps(args), module=module)
        if created:
            task.projects = projects
        else:
            task.projects += projects

        task.priority = priority
        return task

    def complete(self):
        '''Sets the Task state to 2 and populates the finished field.'''
        self.state = 2
        self.finished = datetime.now()

    def __str__(self):
        return '%s (%s: %s)' % (self.module, self.entry, self.entry.path)

    @property
    def jobs(self):
        '''List of jobs related to the task.'''
        return self.job_set.all()

    @property
    def errors(self):
        '''List of errors encountered by related calculations.'''
        return self.entry.errors

    def get_jobs(self, project=None, allocation=None, account=None):
        '''
        Calls the task's entry's "do" method, with first argument as the tasks
        module, and subsequent arguments as the args list. If there is nothing
        left to do, returns an empty sequence.
        '''
        if project is None:
            project = self.get_project()
            while not project:
                print '   -not project'
                time.sleep(30)
                project = self.get_project()
        if allocation is None:
            allocation = project.get_allocation()
            while not allocation:
                print '   -not allocation'
                time.sleep(30)
                allocation = project.get_allocation()
        if account is None:
            account = allocation.get_account(users=list(project.users.all()))

        calc = self.entry.do(self.module, args=json.loads(self.args))
        jobs = []
        #for calc in calcs:
        if calc.instructions:
            self.state = 1
            new_job = Job.create(
                task=self, 
                allocation=allocation, 
                account=account,
                entry=self.entry,
                **calc.instructions)
            jobs.append(new_job)
            calc.save()
        elif calc.converged:
            self.complete()
        else:
            self.state = -1
        return jobs

class Job(models.Model):
    '''
    Job model. Corresponds to a calculation on a compute cluster.

    Job codes:
        -1  :   error
        0   :   eeady to submit
        1   :   running
        2   :   done

    Database attributes:
        qid
        walltime
        path
        run_path
        ncpus

        task
        account
        allocation
        entry

        created
        finished
    '''
    qid = models.IntegerField(default=0)
    walltime = models.DateTimeField(blank=True)
    path = models.CharField(max_length=200)
    run_path = models.CharField(max_length=200)
    ncpus = models.IntegerField(blank=True)
    created = models.DateTimeField(blank=True, auto_now_add=True)
    finished = models.DateTimeField(blank=True, null=True)
    state = models.IntegerField(default=0)

    task = models.ForeignKey(Task)
    entry = models.ForeignKey('Entry')
    account = models.ForeignKey(Account)
    allocation = models.ForeignKey(Allocation)

    class Meta:
        app_label = 'qmpy'
        db_table = 'jobs'

    @staticmethod
    def create(task=None, allocation=None, entry=None,
            account=None,
            path=None, 
            walltime=3600, serial=None,
            header=None,
            mpi=None, binary=None, pipes=None,
            footer=None):

        if entry is None:
            entry = task.entry

        assert isinstance(allocation, Allocation)
        assert isinstance(task, Task)
        assert path is not None

        if account is None:
            account = allocation.get_account()
        
        job = Job(path=path, walltime=walltime, 
                allocation=allocation,
                account=account,
                entry=entry,
                task=task)

        #if walltime < 3600:
        #    nodes = 1
        #    ppn = int(walltime/3600.*job.account.host.ppn)
        #    walltime = walltime/ppn
        #else:
        #    ppn = job.account.host.ppn
        #    nodes = 1+int(walltime/float(job.account.host.walltime))
        #    walltime = walltime/float(ppn*nodes)
        
        if serial:
            ppn = 1
            nodes = 1
            walltime = 3600*24
        else:
            nodes = 1
            ppn = job.account.host.ppn
            walltime = job.account.host.walltime
            
        binary = job.account.host.get_binary(binary)
        if not binary:
            raise AllocationError

        sec = timedelta(seconds=walltime)
        d = datetime(1,1,1) + sec
        job.walltime = d
        walltime = '%02d:%02d:%02d:%02d' % (
                d.day-1, 
                d.hour, 
                d.minute,
                d.second)

        qp = qmpy.INSTALL_PATH + '/configuration/qfiles/'
        text = open(qp+job.account.host.sub_text+'.q', 'r').read()
        qfile = text.format(
                host=allocation.host.name,
                key=allocation.key, name=job.description,
                walltime=walltime, nodes=nodes, ppn=ppn,
                header=header,
                mpi=mpi, binary=binary, pipes=pipes,
                footer=footer)

        qf=open(job.path+'/auto.q', 'w')
        qf.write(qfile)
        qf.close()
        job.ncpus = ppn*nodes
        job.run_path = job.account.run_path+'/'+job.description
        return job

    @property
    def walltime_expired(self):
        from datetime import datetime, timedelta
        elapsed = datetime.now() - self.created
        if elapsed.total_seconds() > self.walltime:
            return True
        else:
            return False

    @property
    def subdir(self):
        return self.path.replace(self.entry.path, '')

    @property
    def description(self):
        return '{entry}_{subdir}_{uniq}'.format(
                entry=self.entry.id,
                subdir=self.subdir.strip('/').replace('/','_'),
                uniq='-'.join(json.loads(self.task.args)))

    def __str__(self):
        return '%s on %s' % (self.description, self.account)

    def is_done(self):
        if datetime.now() + timedelta(seconds=-600) > self.created:
            return False

        return not ( self.qid in self.account.host.running_now )

    def submit(self, verbosity=0):
        self.created = datetime.now()
        self.qid = self.account.submit(path=self.path,
                run_path=self.run_path,
                qfile='auto.q', verbosity=verbosity)
        self.task.state = 1
        self.state = 1

    def collect(self, verbosity=0):
        if self.account.host.state == -1:
            return
        self.task.state = 0
        self.task.save()
        self.state = 2
        self.account.copy(move=True,
                to='local', destination=self.path,
                folder=self.run_path, file='*')
        self.account.execute('rm -rf %s' % self.run_path, ignore_output=True)
        self.finished = datetime.now()




# gclp.space

import numpy as np
import numpy.linalg
import json
import itertools
from collections import defaultdict
from utils import *
from phase import Phase

from scipy.spatial import ConvexHull
from pulp import LpProblem,lpSum,LpVariable,LpMinimize,value

class Reaction(object):
    __doc__ = '''
    methods
    attributes
    '''
    
    def __init__(self, products=None, reactants=None, delta_var=None,
            variable=None, electrons=1.0):
        ### write blanks
        self._products = products
        self._reactants = reactants
        if isinstance(variable, basestring):
            self._variable = parse_comp(variable)
        elif isinstance(variable, Phase):
            self._variable = variable.unit_comp
        else:
            self._variable = variable
        self._delta_var = delta_var
        self._delta_h = None
        self.electrons = electrons
        self._base = None

    def __repr__(self):
        return '<Reaction %s + %s --> %s : %s>' % (self.reactant_string,
                self.var_string, self.product_string,
                self.voltage)

    #### attributes

    @property
    def p_var_comp(self):
        if self.p_natoms == 0:
            return 1.0
        vamt = sum( p.fraction(self._variable)['var']*amt for p, amt in
                self._products.items())*sum(self.variable.values())
        return vamt/self.p_natoms

    @property
    def p_natoms(self):
        return sum( sum(p.unit_comp.values())*amt for p, amt in 
                self._products.items())

    @property
    def r_var_comp(self):
        if self.r_natoms == 0:
            return 1.0
        vamt = sum( p.fraction(self._variable)['var']*amt for p, amt in
                self._reactants.items())*sum(self.variable.values())
        return vamt/self.r_natoms

    @property
    def r_bgap(self):
        if not self._reactants:
            return 0.0
        else:
            return min([ p.band_gap for p in self._reactants if p.band_gap is not None])

    @property
    def p_bgap(self):
        if not self._reactants:
            return 0.0
        else:
            return min([ p.band_gap for p in self._products if p.band_gap is not None])

    @property
    def r_natoms(self):
        return sum( sum(p.unit_comp.values())*amt for p, amt in 
                self._reactants.items())

    @property
    def base(self):
        if self._base is None:
            self._base = min([
                min(amt/sum(phase.nom_comp.values()) for phase, amt in
                    self._products.items() ),
                min(amt/sum(phase.nom_comp.values()) for phase, amt in
                    self._reactants.items())])
        return self._base

    @property
    def product_string(self):
        return ' + '.join( '%s %s' % 
                (amt/sum(phase.nom_comp.values())/self.base, phase.name) 
                for phase, amt in self._products.items() )

    @property
    def reactant_string(self):
        return ' + '.join( '%s %s' %
                (amt/sum(phase.nom_comp.values())/self.base, phase.name) 
                for phase, amt in self._reactants.items() )

    @property
    def product_latex(self):
        return ' + '.join( '%s %s' %
                (amt/sum(phase.nom_comp.values())/self.base, phase.latex) 
                for phase, amt in self._products.items() )

    @property
    def reactant_latex(self):
        return ' + '.join( '%s %s' %
                (amt/sum(phase.nom_comp.values())/self.base, phase.latex) 
                for phase, amt in self._reactants.items() )

    @property
    def var_string(self):
        return '%s %s' % (self.delta_var/self.base, format_comp(self.variable,
            special='unit'))

    @property
    def var_latex(self):
        return '%s %s' % (self.delta_var, comp_to_latex(self.variable,
            special='reduce'))

    @property
    def react_mass(self):
        if not self._reactants:
            return 0
        return sum( amt*phase.mass for phase, amt in 
                self._reactants.items() ) / sum(self._reactants.values())

    @property
    def prod_mass(self):
        if not self._products:
            return 0
        return sum( amt*phase.mass for phase, amt in 
                self._products.items() ) / sum(self._products.values())

    @property
    def react_vol(self):
        if not self._reactants:
            return 0
        return sum( amt*phase.volume for phase, amt in 
                self._reactants.items() ) / sum(self._reactants.values())

    @property
    def prod_vol(self):
        if not self._products:
            return 0
        return sum( amt*phase.volume for phase, amt in 
                self._products.items() ) / sum(self._products.values())

    @property
    def delta_h(self):
        if self._delta_h is None:
            self.get_delta_h()
        return self._delta_h

    @property
    def delta_var(self):
        if self._delta_var is None:
            self._delta_var = self.p_var_comp - self.r_var_comp
            if self._delta_var == 0:
                self._delta_var = 1 
        return self._delta_var

    @property
    def variable(self):
        if self._variable is None:
            self.get_variable()
        return self._variable

    #### methods

    def get_delta_h(self):
        re = 0.0
        for k, v in self._reactants.items():
            re += v*k.energy
        pe = 0.0
        for k, v in self._products.items():
            pe += v*k.energy
        self._delta_h = re - pe

    @property
    def voltage(self):
        return self.delta_h/self.delta_var/self.electrons



from django.db import models

import qmpy.materials.element as elt

class Potential(models.Model):
    potid = models.AutoField(primary_key=True)
    potcar = models.TextField()
    element = models.ForeignKey(elt.Element)

    name = models.CharField(max_length=10)
    xc = models.CharField(max_length=3)
    gw = models.BooleanField(default=False)
    paw = models.BooleanField(default=False)
    us = models.BooleanField(default=False)
    enmax = models.FloatField()
    enmin = models.FloatField()
    date = models.CharField(max_length=20)
    electrons = models.TextField(blank=True, null=True)

    class Meta:
        app_label = 'qmpy'
        db_table = 'vasp_potentials'

    def __str__(self):
        ident = '%s %s' % (self.name, self.xc)
        if self.paw:
            ident += ' PAW'
        if self.us:
            ident += ' US'
        if self.gw:
            ident += ' GW'
        return ident

    @classmethod
    def read_potcar(cls, potfile):
        pots = open(potfile).read()
        pots = pots.strip().split('End of Dataset')
        potobjs = []
        for pot in pots:
            if not pot:
                continue
            potcar = {}
            for line in pot.split('\n'):
                if 'TITEL' in line:
                    potcar['name'] = line.split()[3]
                    telt = potcar['name'].split('_')[0]
                    date = potcar['name'].split('_')[-1]
                    potcar['element'] = elt.Element.objects.get(symbol=telt)
                    if 'GW' in line:
                        potcar['gw'] = True
                    if 'PAW' in line:
                        potcar['paw'] = True
                    if 'US' in line:
                        potcar['us'] = True
                if 'ENMAX' in line:
                    data = line.split()
                    potcar['enmax'] = float(data[2].rstrip(';'))
                    potcar['enmin'] = float(data[5])
                if 'VRHFIN' in line:
                    potcar['electrons'] = line.split(':')[1]
                if 'LEXCH' in line:
                    key = line.split()[-1]
                    if key == '91':
                        potcar['xc'] = 'GGA'
                    elif key == 'CA':
                        potcar['xc'] = 'LDA'
                    elif key == 'PE':
                        potcar['xc'] = 'PBE'
            potobj, created = cls.objects.get_or_create(**potcar)
            if created:
                potobj.potcar = pot
            potobjs.append(potobj)
        return potobjs

class Hubbard(models.Model):

    ## specifications
    element = models.ForeignKey(elt.Element, related_name='hubbards')
    convention = models.CharField(max_length=20)
    ox = models.FloatField(default=None, null=True)
    ligand = models.ForeignKey(elt.Element, related_name='+',
            null=True, blank=True)

    ## parameters
    u = models.FloatField(default=0)
    l = models.IntegerField(default=-1)
    correction = models.FloatField(default=0.0)

    class Meta:
        app_label = 'qmpy'
        db_table = 'hubbards'

    def __nonzero__(self):
        if self.u > 0 and self.l != -1:
            return True
        else:
            return False

    def __eq__(self, other):
        if self.element != other.element:
            return False
        elif self.ligand != other.ligand:
            return False
        elif self.u != other.u:
            return False
        elif self.l != other.l:
            return False
        return True

    def __str__(self):
        retval = self.element_id
        if self.ox:
            retval += '+%d' % (self.ox)
        if self.ligand:
            retval += '-'+self.ligand_id
        retval += ', U=%0.2f, L=%d' % (self.u, self.l)
        return retval

    @property
    def key(self):
        return '%s_%s' % (self.element_id, self.u)

    @classmethod
    def get(cls, elt, ox=None, u=0, l=-1, lig=None):
        hub, new = Hubbard.objects.get_or_create(
                element_id=elt,
                ligand=lig, ox=ox,
                l=l, u=u)
        if new:
            hub.save()
        return hub



#!/usr/bin/env python

'''
Interface to ASE Atoms, and Calculators.
'''

from qmpy.materials.structure import Structure
from qmpy.materials.atom import Atom

#try:
import ase
import ase.io
ASE_INSTALLED = True
#except ImportError:
#    ASE_INSTALLED = False
#    pass

def structure_to_atoms(structure):
    if not ASE_INSTALLED:
        print 'ASE must be installed to convert a Structure to an Atoms object'
        return
    atoms = Atoms(cell=structure.cell,
            scaled_positions=structure.scaled_coords, 
            magmoms=structure.magmoms,
            symbols=''.join(sorted(a.symbol for a in structure.atoms)))
    return atoms

def atoms_to_structure(atoms):
    if not ASE_INSTALLED:
        print 'ASE must be installed to convert Atoms object to a Structure'
        return
    struct = Structure()
    struct.cell = atoms.get_cell()
    for a in atoms: 
        atom = Atom()
        atom.symbol = a.symbol
        atom.coord = a.position
        atom.magmom = a.magmom
        atom.direct = False
        struct.add_atom(atom)
    return struct

def read_ase(filename):
    if not ASE_INSTALLED:
        print 'ASE must be installed to convert Atoms object to a Structure'
        return

    atoms = ase.io.read(filename)
    return atoms_to_structure(atoms)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
import time
import os

from django.db import models

from qmpy.db.custom import *
from qmpy.materials.composition import *
from qmpy.materials.element import Element, Species
from qmpy.materials.structure import Structure
from qmpy.utils import *
from qmpy.computing.resources import Project
from qmpy.data.meta_data import *
import qmpy.io as io
import qmpy.computing.scripts as scripts

k_desc = 'Descriptive keyword for looking up entries'
h_desc = 'A note indicating a reason the entry should not be calculated'

@add_meta_data('keyword', description=k_desc)
@add_meta_data('hold', description=h_desc)
class Entry(models.Model):
    '''
    The core model for typical database entries. An Entry model represents an
    input structure to the database, and can be created from any input file.
    
    Database fields:
        path                : Path to the Entry. (string)
        meta_data           : QuerySet of MetaData models. MetaData are tags
                              consisting of a type and a value. e.g.
                              MetaData(type="hold", value="partially occupied")
        label               : Brief description of the entry, e.g. ICSD-1415
                              (string)
        duplicate_of        : If this entry appears to be an exact duplicate of
                              another existing entry, the existing entry is
                              linked here.
        duplicates          : If another entry duplicates this one, you can
                              find those duplicate entries throug this
                              QuerySet.
        reference           : Journal reference for the input structure 
                              (Reference)
        prototype           : If this structure has been assigned to be an
                              instance of a prototype, it is specified here
        composition         : Composition object matching the composition of
                              input structure


        stable              : Is the structure on the convex hull?
        delta_e             : Formation energy per atom of structure (None if
                              not calculated) (float)

        element_set         : QuerySet of associated Elements
        species_set         : QuerySet of assicated Speices
        project_set         : QuerySet of associated Projects
        structure_set       : QuerySet of associated structures. Easier to
                              access this information through the structures
                              attribute.
        calculation_set     : QuerySet of associated calculations. Easier to
                              access this information through the calculations
                              attribute.
        job_set
        task_set

    '''
    ### structure properties
    path = models.CharField(max_length=255, db_index=True, unique=True)
    meta_data = models.ManyToManyField('MetaData')
    label = models.CharField(max_length=20, null=True)

    ### record keeping
    duplicate_of = models.ForeignKey('Entry', related_name='duplicates',
            null=True)

    ### 
    stable = models.BooleanField(default=False)
    delta_e = models.FloatField(blank=True, null=True)

    ## links
    element_set = models.ManyToManyField('Element')
    species_set = models.ManyToManyField('Species')
    project_set = models.ManyToManyField('Project')
    composition = models.ForeignKey('Composition', blank=True, null=True)
    spacegroup = models.ForeignKey('Spacegroup', blank=True, null=True)
    reference = models.ForeignKey('Reference', null=True, blank=True)
    prototype = models.ForeignKey('Prototype', null=True, blank=True)

    class Meta:
        app_label = 'qmpy'
        db_table = 'entries'

    def __str__(self):
        return '%s : %s' % (self.id, self.name)

    def save(self, *args, **kwargs):
        '''Saves the Entry, as well as all associated objects.'''
        if not self.reference is None:
            if self.reference.id is None:
                self.reference.save()
                self.reference = self.reference
        self.delta_e = self.energy
        super(Entry, self).save(*args, **kwargs)
        for k, v in self.structures.items():
            if v.label != k:
                v.label = k
                v.entry = self
                v.save()
        for k, v in self.calculations.items():
            if v.label != k:
                v.label = k
                v.entry = self
                v.save()
        self.element_set = self.elements
        self.species_set = self.species
        self.project_set = self.projects
        self.meta_data = self.hold_objects + self.keyword_objects

    @classmethod
    def create(cls, source, keywords=[], projects=[], **kwargs):
        '''
        Create an Entry object from a provided input file.

        Optional arguments:
            keywords - list of keywords to associate with the entry
            projects - list of projects to associate with the entry

        '''
        source_file = os.path.abspath(source)
        path = os.path.dirname(source_file)
        if Entry.objects.filter(path=path).exists():
            return Entry.objects.get(path=path)
        entry = cls(**kwargs)
        
        entry.source_file = source_file
        entry.path = os.path.dirname(source_file)
        structure = io.read(source_file)
        structure.make_primitive()

        entry.structures['input'] = structure
        if any([ a.occupancy < 0.95 for a in structure ]):
            entry.holds.append('partial occupancy')
        if not structure.composition == structure.reported_composition:
            entry.holds.append('composition mismatch in CIF')
            entry.composition = entry.input.reported_composition
        else:
            entry.composition = entry.input.composition

        entry.ntypes = structure.ntypes
        entry.natoms = len(structure.atoms)
        entry.elements = entry.comp.keys() 
        entry.reference = structure.reference
        entry.prototype = structure.prototype
        entry.keywords = keywords
        entry.projects = projects
        return entry

    _elements = None
    @property
    def elements(self):
        '''List of Elements'''
        if self._elements is None:
            self._elements = [ Element.get(e) for e in self.comp.keys() ]
        return self._elements

    @elements.setter
    def elements(self, elements):
        self._elements = [ Element.get(e) for e in elements ]

    _species = None
    @property
    def species(self):
        '''List of Species'''
        if self._species is None:
            self._species = [ Species.get(s) for s in self.spec_comp.keys() ]
        return self._species

    @species.setter
    def species(self, species):
        self._species = [ Species.get(e) for e in species ]

    _projects = None
    @property
    def projects(self):
        '''List of Projects'''
        if self._projects is None:
            self._projects = list(self.project_set.all())
        return self._projects

    @projects.setter
    def projects(self, projects):
        self._projects = [ Project.get(p) for p in projects ]

    _structures = None
    @property
    def structures(self):
        if self._structures is None:
            if self.id is None:
                self._structures = {}
            else:
                structs = self.structure_set.exclude(label='')
                labels = [ s.label for s in structs ]
                if len(set(labels)) < len(labels):
                    print 'Warning! Some related objects have repeat names!'
                    print 'To address this, all labels have been appended',
                    print 'with the primary key of the related object.'
                    labels = [ '%s_%s' % (s.label, s.pk) for s in structs ]
                self._structures = dict((k, s) for k,s in zip(labels, structs))
        return self._structures

    @structures.setter
    def structures(self, structs):
        if not isinstance(structs, dict):
            raise TypeError('structures must be a dict')
        if not all( isinstance(v, Structure) for v in structs.values()):
            raise TypeError('structures must be a dict of Calculations')
        self._structures = structs

    _calculations = None
    @property
    def calculations(self):
        '''Dictionary of label:Calculation pairs.''' 
        if self._calculations is None:
            if self.id is None:
                self._calculations = {}
            else:
                calcs = self.calculation_set.exclude(label='')
                labels = [ c.label for c in calcs ]
                if len(set(labels)) < len(labels):
                    print 'Warning! Some related objects have repeat names!'
                    print 'To address this, all labels have been appended',
                    print 'with the primary key of the related object.'
                    labels = [ '%s_%s' % (c.label, c.pk) for c in calcs ]
                self._calculations = dict((k, c) for k,c in zip(labels, calcs))
        return self._calculations

    @calculations.setter
    def calculations(self, calcs):
        if not isinstance(calcs, dict):
            raise TypeError('calculations must be a dict')
        if not all( isinstance(v, Calculation) for v in calcs.values()):
            raise TypeError('calculations must be a dict of Calculations')
        self._calculations = calcs

    @property
    def input(self):
        return self.structures.get('input')

    @input.setter
    def input(self, structure):
        self.structures['input'] = structure

    @property
    def tasks(self):
        return list(self.task_set.all())

    @property
    def jobs(self):
        return list(self.job_set.all())

    @property
    def comp(self):
        if not self.composition_id is None:
            return parse_comp(self.composition_id)
        elif not self.input is None:
            return self.input.comp
        else:
            return {}

    @property
    def spec_comp(self):
        '''
        Composition dictionary, using species (element + oxidation state)
        instead of just the elements.
        
        '''
        if self.input is None:
            return {}
        else:
            return self.input.spec_comp

    @property
    def unit_comp(self):
        '''Composition dictionary, normalized to 1 atom.'''
        return normalize_comp(self.comp)

    @property
    def red_comp(self):
        '''Composition dictionary, in reduced form.'''
        return reduce_comp(self.comp)

    @property
    def name(self):
        '''Unformatted name'''
        return format_comp(reduce_comp(self.comp))

    @property
    def latex(self):
        '''LaTeX formatted name'''
        return latex_comp(reduce_comp(self.comp))

    @property
    def html(self):
        '''HTML formatted name'''
        return html_comp(reduce_comp(self.comp))

    @property
    def space(self):
        '''Return the set of elements in the input structure.
        
        Example:
        >>> e = Entry.create("fe2o3/POSCAR") # an input containing Fe2O3
        >>> e.space
        set(["Fe", "O"])
        
        '''
        return set([ e.symbol for e in self.elements])

    @property
    def total_energy(self):
        '''
        If the structure has been relaxed, returns the formation energy of the
        final relaxed structure. Otherwise, returns None.

        '''
        if 'static' in self.calculations:
            if self.calculations['static'].converged:
                return self.calculations['static'].energy_pa
        if 'standard' in self.calculations:
            if self.calculations['standard'].converged:
                return self.calculations['standard'].energy_pa
        return None

    @property
    def energy(self):
        '''
        If the structure has been relaxed, returns the formation energy of the
        final relaxed structure. Otherwise, returns None.

        '''
        if 'static' in self.calculations:
            if self.calculations['static'].converged:
                return self.calculations['static'].delta_e
        if 'standard' in self.calculations:
            if self.calculations['standard'].converged:
                return self.calculations['standard'].delta_e
        return None

    #@property
    #def spacegroup(self):
    #    for key in ['relaxed', 'fine_relaxed', 'input']:
    #        if key in self.structures:
    #            return self.structures[key].spacegroup

    @property
    def mass(self):
        '''Return the mass of the entry, normalized to per atom.'''
        return sum( Element.objects.get(symbol=elt).mass*amt for 
                elt, amt in self.unit_comp)

    @property
    def volume(self):
        '''
        If the entry has gone through relaxation, returns the relaxed
        volume. Otherwise, returns the input volume.
        
        '''
        if not self.relaxed is None:
            return self.relaxed.volume/self.natoms
        else:
            return self.input.volume/self.natoms

    @property
    def errors(self):
        '''List of errors encountered in all calculations.'''
        return dict( ( c.path, c.get_errors()) for c in self.calculations )

    @property
    def chg(self):
        '''
        Attempts to load the charge density of the final calculation, if it is
        done. If not, returns False.

        '''
        if not hasattr(self, '_chg'):
            if not self.done:
                self._chg = False
            else:
                self._chg = Grid.load_xdensity(self.path+'/standard/CHGCAR.gz')
        return self._chg

    def do(self, module, *args, **kwargs):
        '''
        Looks for a computing script matching the first argument, and attempts
        to run it with itself as the first argument. Sends *args and **kwargs
        to the script. Should return a Calculation object, or list of
        Calculation objects. 

        Example:
        >>> e = Entry.objects.get(id=123)
        >>> e.do('relaxation')
        <Calculation: 523 @ relaxation settings>

        '''
        script = getattr(scripts, module)
        return script(self, *args, **kwargs)

    def move(self, path):
        '''
        Moves all calculation files to the specified path.

        '''
        path = os.path.abspath(path)
        try:
            os.system('mv %s %s' % (self.path, path))
        except Exception, err:
            print err
            return
        self.path = path
        print 'Moved %s to %s' % (self, path)
        self.save()

    @property
    def running(self):
        return self.job_set.filter(state=1)

    def reset(self):
        '''
        Deletes all calculations, removes all associated structures - returns
        the entry to a pristine state.

        '''
        for structure in self.structures:
            if structure.label != 'source':
                structure.delete()
        
        for calc in self.calculations:
            calc.clean_start()
            calc.delete()

        for task in self.tasks:
            task.state = 0 
            task.save()

        for job in self.job_set.filter(state=1):
            job.collect()
            job.delete()
        self.job_set.all().delete()

        for dir in ['initialize', 'coarse_relax', 'fine_relax', 'standard']:
            os.system('rm -rf %s/%s &> /dev/null' % (self.path, dir))

    def visualize(self, structure='source'):
        '''Attempts to open the input structure for visualization using VESTA'''
        os.system('VESTA %s/POSCAR' % self.path)




#/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from os.path import dirname, abspath
import sys

### Kludge to get the django settings module into the path
INSTALL_PATH = dirname(__file__)
sys.path.insert(-1, INSTALL_PATH)
if 'DJANGO_SETTINGS_MODULE' not in os.environ:
    os.environ['DJANGO_SETTINGS_MODULE'] = 'qmpy.db.settings'

from models import *



from setuptools import setup

setup(
    name='qmpy',
    version='0.3.1',
    author='S. Kirklin',
    author_email='scott.kirklin@gmail.com',
    packages=['qmpy', 
        'qmpy.analysis', 
        'qmpy.io', 
        'qmpy.db',
        'qmpy.data',
        'qmpy.materials',
        'qmpy.analysis',
        'qmpy.analysis.vasp',
        'qmpy.analysis.thermodynamics',
        'qmpy.analysis.symmetry',
        'qmpy.computing',
        'qmpy.configuration', 
        'qmpy.utils'],
    scripts=['bin/oqmd'],
    url='http://pypi.python.org/pypi/qmpy',
    license='LICENSE.txt',
    description='Suite of computational materials science tools',
    long_description=open('README.txt').read(),
    install_requires=[
        "Django >= 1.1.1",
        "PuLP",
        "numpy >= 1.6.4",
        "scipy >= 0.12.0",
        "PyYAML",
        "MySQL-python",
        "pytest",
        "PyCifRW",
        "matplotlib",
        "networkx",
        "python-memcached",
        "django-extensions",
        "argparse"
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from django.db import models

class MetaData(models.Model):
    '''
    Model for arbitrary meta-data descriptors for various qmpy objects.
    Generally accessed by properties and methods added by the "add_label"
    descriptor. See "add_label" for a more detailed description of its use

    Example:
    >>> MetaData.get('Keyword', 'ICSD')
    <Keyword: ICSD>
    '''
    type = models.CharField(max_length=15)
    value = models.CharField(max_length=63)

    class Meta:
        app_label = 'qmpy'
        db_table = 'meta_data'

    # builtins
    def __str__(self):
        return self.value

    def __repr__(self):
        return '<%s: %s>' % (self.type.title(), self.value)

    # accessor
    @classmethod
    def get(cls, type, value):
        md, new = MetaData.objects.get_or_create(type=type, value=value)
        if new:
            md.save()
        return md

def add_meta_data(label, plural=None, cache=None, description=''):
    '''
    Decorator for making easily creatable and accessible typed labels for
    django models.

    Requires that the class being decorated has a many_to_many field with
    MetaData.

    Example:
    >>> @add_label("keywords")
    >>> class NewModel(models.Model):
    >>>     meta_data = models.ManyToManyField('MetaData')
    >>> 
    >>> instance = NewModel()
    >>> instance.keywords
    []
    >>> instance.add_keyword('decorated!')
    >>> instance.keywords
    [<Keyword: decorated!>]
    >>> instance.remove_keyword('decorated!')
    >>> instance.keywords
    []
    >>> instance.keywords = ['add', 'in', 'bulk']
    >>> instance.keywords
    [<Keyword: add>, <Keyword: in>, <Keyword: bulk>]

    '''
    if plural is None:
        plural = '%ss' % label
    if cache is None:
        cache = '_%s' % plural
    label = label.lower()
    plural = plural.lower()
    cache = cache.lower()

    def func(cls):
        setattr(cls, cache, None)
        def getter(self):
            if getattr(self, cache) is None:
                if getattr(self, 'pk') is None:
                    setattr(self, cache, [])
                else:
                    data = self.meta_data.filter(type=label)
                    data = data.values_list('value', flat=True)
                    setattr(self, cache, list(data))
            return getattr(self, cache)

        def setter(cls, values):
            setattr(cls, cache, values)

        setattr(cls, plural, property(getter, 
                                     setter, 
                                     None, 
                                     description))
        def adder(cls, value):
            'Helper function to add %s to list of %s.' % (label, plural)
            existing = getattr(cls, plural)
            existing.append(value)
        setattr(cls, 'add_%s' % label, adder)

        def remover(cls, value):
            'Helper function to remove a %s from list of %s.' % (label, plural)
            existing = getattr(cls, plural)
            if value in existing:
                existing.remove(value)
        setattr(cls, 'remove_%s' % label, remover)

        doc = 'Return list of %s' % plural
        def obj_getter(self):
            return [ MetaData.get(label, v) for v in getattr(self, plural) ]
        setattr(cls, '%s_objects' % label, property(obj_getter, None, None, doc))

        return cls
    return func



import yaml
import os

from qmpy import INSTALL_PATH
from qmpy.configuration import VASP_POTENTIALS, sync_resources
from qmpy.models import *

sync_resources()
data = open(INSTALL_PATH+'/data/spacegroups.yml').read()
Spacegroup.objects.all().delete()
spacegroups = yaml.load(data)
for sgd in spacegroups.values():
    sg = Spacegroup(number=sgd['number'],
                    hm=sgd['hm'],
                    hall=sgd['hall'],
                    schoenflies=sgd['schoenflies'],
                    lattice_system=sgd['system'])
    sg.save()
    print sg.number
    cvs = []
    for cv in sgd['centering_vectors']:
        cvs.append(Translation.get(cv))
    sg.centering_vectors = cvs

    ops = []
    for op in sgd['sym_ops']:
        ops.append(Operation.get(','.join(op)))
    sg.sym_ops = ops
    sg.save()

    wycks = []
    for k, site in sgd['wyckoff_sites'].items():
        wycks.append(WyckoffSite(symbol=k, 
                           x=site['coordinate'].split()[0],
                           y=site['coordinate'].split()[1],
                           z=site['coordinate'].split()[2],
                           multiplicity=site['multiplicity']))
    sg.site_set = wycks

elements = open(INSTALL_PATH+'/data/elements/data.yml').read()
Element.objects.all().delete()

for elt, data in yaml.load(elements).items():
    e = Element(**data)
    e.save()

hubs = open(INSTALL_PATH+'/configuration/vasp_settings/hubbards.yml').read()

for group, hubbard in yaml.load(hubs).items():
    for ident, data in hubbard.items():
        elt, ligand, ox = ident.split('_')
        hub = Hubbard(
                l=data['L'],
                u=data['U'])
        if ox != '*':
            hub.ox = ox
        hub.ligand_id = ligand
        hub.element_id = elt
        hub.convention = group
        hub.save()

loaded = []
for (path, dirs, files) in os.walk(VASP_POTENTIALS):
    for f in files:
        if 'GW' in path:
            continue
        if 'POTCAR' in files and not path in loaded:
            loaded.append(path)
            try:
                pots = Potential.read_potcar(path+'/POTCAR')
                for pot in pots:
                    pot.save()
            except Exception, err:
                print 'Couldn\'t load:', path

